import pytest

from leapp import reporting
from leapp.libraries.actor import checkinstalledkernels
from leapp.libraries.common.config import architecture
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked, logger_mocked
from leapp.libraries.stdlib import api
from leapp.models import RPM, InstalledRedHatSignedRPM

RH_PACKAGER = 'Red Hat, Inc. <http://bugzilla.redhat.com/bugzilla>'

# Do not make sense to run any tests when the module is not accessible
pytest.importorskip("rpm")


def create_rpm(
            version,
            release,
            name='kernel',
            packager=RH_PACKAGER,
            pgpsig='SOME_OTHER_SIG_X',
            epoch='0',
            ):
    return RPM(
        name=name,
        arch=release.split('.')[-1],
        version=version,
        release='.'.join(release.split('.')[0:-1]),
        epoch='0',
        packager=RH_PACKAGER,
        pgpsig='SOME_OTHER_SIG_X',
    )


def create_rpms(pkgs):
    installed_rpms = InstalledRedHatSignedRPM()
    for pkg in pkgs:
        installed_rpms.items.append(
            create_rpm(name=pkg[0], version=pkg[1], release=pkg[2]))
    return installed_rpms


@pytest.mark.parametrize('vra,version,release', [
    ('3.10.0-1234.21.1.el7.x86_64', '3.10.0', '1234.21.1.el7.x86_64'),
    ('5.8.8-100.fc31.x86_64', '5.8.8', '100.fc31.x86_64'),
])
def test_current_kernel(monkeypatch, vra, version, release):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(kernel=vra))
    assert version == checkinstalledkernels.get_current_kernel_version()
    assert release == checkinstalledkernels.get_current_kernel_release()


s390x_pkgs_single = [
    ('kernel', '3.10.0', '957.43.1.el7.s390x'),
    ('something', '3.10.0', '957.43.1.el7.s390x'),
    ('kernel-something', '3.10.0', '957.43.1.el7.s390x')
]
s390x_pkgs_multi = [
    ('kernel', '3.10.0', '957.43.1.el7.s390x'),
    ('something', '3.10.0', '957.43.1.el7.s390x'),
    ('kernel', '3.10.0', '956.43.1.el7.s390x')
]


def test_single_kernel_s390x(monkeypatch):
    msgs = [create_rpms(s390x_pkgs_single)]
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(
        arch=architecture.ARCH_S390X,
        msgs=msgs,
        kernel='3.10.0-957.43.1.el7.s390x'),
    )
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    checkinstalledkernels.process()
    assert not reporting.create_report.called


def test_multi_kernel_s390x(monkeypatch):
    msgs = [create_rpms(s390x_pkgs_multi)]
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(
        arch=architecture.ARCH_S390X,
        msgs=msgs,
        kernel='3.10.0-957.43.1.el7.s390x'),
    )
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    checkinstalledkernels.process()
    assert reporting.create_report.called
    assert reporting.create_report.report_fields['title'] == 'Multiple kernels installed'


versioned_kernel_pkgs = [
    ('kernel', '3.10.0', '456.43.1.el7.x86_64'),
    ('kernel', '3.10.0', '789.35.2.el7.x86_64'),
    ('kernel', '3.10.0', '1234.21.1.el7.x86_64')
]


@pytest.mark.parametrize('expect_report,msgs,curr_kernel', [
    (False, [create_rpms(versioned_kernel_pkgs)], '3.10.0-1234.21.1.el7.x86_64'),
    (True, [create_rpms(versioned_kernel_pkgs)], '3.10.0-456.43.1.el7.x86_64'),
    (True, [create_rpms(versioned_kernel_pkgs)], '3.10.0-789.35.2.el7.x86_64'),
])
def test_newest_kernel(monkeypatch, expect_report, msgs, curr_kernel):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(kernel=curr_kernel, msgs=msgs))
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    checkinstalledkernels.process()
    if expect_report:
        assert reporting.create_report.called
        assert reporting.create_report.report_fields['title'] == 'Newest installed kernel not in use'
    else:
        assert not reporting.create_report.called


# put the kernel in the middle of the list so that its position doesn't guarantee its rank
versioned_kernel_pkgs.insert(2, ('kernel', '4.14.0', '115.29.1.el7.x86_64'))


@pytest.mark.parametrize('expect_report,msgs,curr_kernel', [
    (True, [create_rpms(versioned_kernel_pkgs)], '3.10.0-1234.21.1.el7.x86_64'),
    (False, [create_rpms(versioned_kernel_pkgs)], '4.14.0-115.29.1.el7.x86_64'),
])
def test_newest_kernel_more_versions(monkeypatch, expect_report, msgs, curr_kernel):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(kernel=curr_kernel, msgs=msgs))
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    checkinstalledkernels.process()
    if expect_report:
        assert reporting.create_report.called
        assert reporting.create_report.report_fields['title'] == 'Newest installed kernel not in use'
    else:
        assert not reporting.create_report.called


@pytest.mark.parametrize('evr', [
    ('', '3.10.0', '1234.21.1.el7.x86_64'),
    ('', '3.10.0', '456.43.1.el7.x86_64'),
    ('', '3.10.0', '1.1.1.1.1.1.1.2.el7x86_64'),
    ('', '4.10.4', '1234.21.1.el7.x86_64'),
    ('', '6.6.6', '1234.56.rt78.el9.x86_64'),
])
def test_get_evr(monkeypatch, evr):
    pkg = create_rpm(version=evr[1], release=evr[2])
    assert checkinstalledkernels.get_EVR(pkg) == evr


versioned_kernel_rt_pkgs = [
    ('kernel-rt', '3.10.0', '789.35.2.rt56.1133.el7.x86_64'),
    ('kernel-rt', '3.10.0', '789.35.2.rt57.1133.el7.x86_64'),
    ('kernel-rt', '3.10.0', '789.35.2.rt101.1133.el7.x86_64'),
    ('kernel-rt', '3.10.0', '790.35.2.rt666.1133.el7.x86_64'),
]


@pytest.mark.parametrize('msgs,num,name', [
    ([create_rpms(versioned_kernel_rt_pkgs)], 4, 'kernel-rt'),
    ([create_rpms(versioned_kernel_rt_pkgs[0:-1])], 3, 'kernel-rt'),
    ([create_rpms(versioned_kernel_rt_pkgs[0:-2])], 2, 'kernel-rt'),
    ([create_rpms(versioned_kernel_rt_pkgs[0:-3])], 1, 'kernel-rt'),
    ([create_rpms(versioned_kernel_rt_pkgs)], 0, 'kernel'),
    ([create_rpms(versioned_kernel_rt_pkgs)], 0, 'smth'),
    ([create_rpms(versioned_kernel_pkgs)], 0, 'kernel-rt'),
    ([create_rpms(versioned_kernel_pkgs + versioned_kernel_rt_pkgs)], 4, 'kernel-rt'),
    ([create_rpms(versioned_kernel_pkgs + versioned_kernel_rt_pkgs)], 4, 'kernel'),
])
def test_get_pkgs(monkeypatch, msgs, num, name):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=msgs))
    pkgs = checkinstalledkernels.get_pkgs(name)
    assert len(pkgs) == num


@pytest.mark.parametrize('expect_report,msgs,curr_kernel', [
    # kernel-rt only
    (True, [create_rpms(versioned_kernel_rt_pkgs)], '3.10.0-789.35.2.rt56.1133.el7.x86_64'),
    (True, [create_rpms(versioned_kernel_rt_pkgs)], '3.10.0-789.35.2.rt57.1133.el7.x86_64'),
    (True, [create_rpms(versioned_kernel_rt_pkgs)], '3.10.0-789.35.2.rt101.1133.el7.x86_64'),
    (False, [create_rpms(versioned_kernel_rt_pkgs)], '3.10.0-790.35.2.rt666.1133.el7.x86_64'),
    (False, [create_rpms(versioned_kernel_rt_pkgs[0:-1])], '3.10.0-789.35.2.rt101.1133.el7.x86_64'),
    (False, [create_rpms(versioned_kernel_rt_pkgs[0:1])], '3.10.0-789.35.2.rt56.1133.el7.x86_64'),

    # mix of kernel-rt + kernel
    (
        True,
        [create_rpms(versioned_kernel_rt_pkgs + versioned_kernel_pkgs)],
        '3.10.0-789.35.2.rt101.1133.el7.x86_64'
    ),
    (
        False,
        [create_rpms(versioned_kernel_rt_pkgs + versioned_kernel_pkgs)],
        '3.10.0-790.35.2.rt666.1133.el7.x86_64'
    ),
    (
        True,
        [create_rpms(versioned_kernel_rt_pkgs + versioned_kernel_pkgs)],
        '3.10.0-1234.21.1.el7.x86_64'
    ),
    (
        False,
        [create_rpms(versioned_kernel_rt_pkgs + versioned_kernel_pkgs)],
        '4.14.0-115.29.1.el7.x86_64'
    ),
])
def test_newest_kernel_realtime(monkeypatch, expect_report, msgs, curr_kernel):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(kernel=curr_kernel, msgs=msgs))
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    checkinstalledkernels.process()
    if expect_report:
        assert reporting.create_report.called
        assert reporting.create_report.report_fields['title'] == 'Newest installed kernel not in use'
    else:
        assert not reporting.create_report.called



from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import InitrdIncludes, InstalledTargetKernelVersion


def process():
    files = [f for i in api.consume(InitrdIncludes) for f in i.files]
    if not files:
        api.current_logger().debug("No additional files required to add into the initrd.")
        return

    target_kernel = next(api.consume(InstalledTargetKernelVersion), None)
    if not target_kernel:
        raise StopActorExecutionError(
            "Cannot get version of the installed RHEL-8 kernel",
            details={'Problem': 'Did not receive a message with installed RHEL-8 kernel version'
                                ' (InstalledTargetKernelVersion)'})

    try:
        # multiple files need to be quoted, see --install in dracut(8)
        cmd = ["dracut", "--install", '"' + " ".join(files) + '"', "-f", "--kver", target_kernel.version]
        run(cmd)
    except CalledProcessError as e:
        # just hypothetic check, it should not die
        raise StopActorExecutionError("Cannot regenerate dracut image.", details={"details": str(e)})



from leapp.actors import Actor
from leapp.libraries.actor import cpu
from leapp.models import CPUInfo, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckCPU(Actor):
    """
    Check whether the CPU is supported by the target system. Inhibit upgrade if not.

    Currently we know just about cases with s390x where the set of CPUs supported
    by RHEL 8 is subset of CPUs supported on RHEL 7. We can detect such cases based
    on the machine field inside the /proc/cpuinfo file. expected values of the
    field on supported machines are: 2964, 2965, 3906, 3907, 8561, 8562.
    """

    name = "checkcpu"
    consumes = (CPUInfo,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag,)

    def process(self):
        cpu.process()



import contextlib
import os
import shutil
from collections import namedtuple

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import mounting, utils
from leapp.libraries.stdlib import CalledProcessError, api, run


OVERLAY_DO_NOT_MOUNT = ('tmpfs', 'devpts', 'sysfs', 'proc', 'cramfs', 'sysv', 'vfat')


MountPoints = namedtuple('MountPoints', ['fs_file', 'fs_vfstype'])


def _ensure_enough_diskimage_space(space_needed, directory):
    stat = os.statvfs(directory)
    if (stat.f_frsize * stat.f_bavail) < (space_needed * 1024 * 1024):
        message = ('Not enough space available for creating required disk images in {directory}. ' +
                   'Needed: {space_needed} MiB').format(space_needed=space_needed, directory=directory)
        api.current_logger().error(message)
        raise StopActorExecutionError(message)


def _get_mountpoints(storage_info):
    mount_points = set()
    for entry in storage_info.fstab:
        if os.path.isdir(entry.fs_file) and entry.fs_vfstype not in OVERLAY_DO_NOT_MOUNT:
            mount_points.add(MountPoints(entry.fs_file, entry.fs_vfstype))
        elif os.path.isdir(entry.fs_file) and entry.fs_vfstype == 'vfat':
            api.current_logger().warning(
                'Ignoring vfat {} filesystem mount during upgrade process'.format(entry.fs_file)
            )

    return list(mount_points)


def _mount_name(mountpoint):
    return 'root{}'.format(mountpoint.replace('/', '_'))


def _mount_dir(mounts_dir, mountpoint):
    return os.path.join(mounts_dir, _mount_name(mountpoint))


def _prepare_required_mounts(scratch_dir, mounts_dir, mount_points, xfs_info):
    result = {
        mount_point.fs_file: mounting.NullMount(
            _mount_dir(mounts_dir, mount_point.fs_file)) for mount_point in mount_points
    }

    if not xfs_info.mountpoints_without_ftype:
        return result

    space_needed = _overlay_disk_size() * len(xfs_info.mountpoints_without_ftype)
    disk_images_directory = os.path.join(scratch_dir, 'diskimages')

    # Ensure we cleanup old disk images before we check for space contraints.
    run(['rm', '-rf', disk_images_directory])
    _create_diskimages_dir(scratch_dir, disk_images_directory)
    _ensure_enough_diskimage_space(space_needed, scratch_dir)

    mount_names = [mount_point.fs_file for mount_point in mount_points]

    for mountpoint in xfs_info.mountpoints_without_ftype:
        if mountpoint in mount_names:
            image = _create_mount_disk_image(disk_images_directory, mountpoint)
            result[mountpoint] = mounting.LoopMount(source=image, target=_mount_dir(mounts_dir, mountpoint))
    return result


@contextlib.contextmanager
def _build_overlay_mount(root_mount, mounts):
    if not root_mount:
        raise StopActorExecutionError('Root mount point has not been prepared for overlayfs.')
    if not mounts:
        yield root_mount
    else:
        current = mounts.keys()[0]
        current_mount = mounts.pop(current)
        name = _mount_name(current)
        with current_mount:
            with mounting.OverlayMount(name=name, source=current, workdir=current_mount.target) as overlay:
                with mounting.BindMount(source=overlay.target,
                                        target=os.path.join(root_mount.target, current.lstrip('/'))):
                    with _build_overlay_mount(root_mount, mounts) as mount:
                        yield mount


def _overlay_disk_size():
    """
    Convenient function to retrieve the overlay disk size
    """
    try:
        env_size = os.getenv('LEAPP_OVL_SIZE', default='2048')
        disk_size = int(env_size)
    except ValueError:
        disk_size = 2048
        api.current_logger().warning(
            'Invalid "LEAPP_OVL_SIZE" environment variable "%s". Setting default "%d" value', env_size, disk_size
        )
    return disk_size


def cleanup_scratch(scratch_dir, mounts_dir):
    """
    Function to cleanup the scratch directory
    """
    api.current_logger().debug('Cleaning up mounts')
    if os.path.ismount(mounts_dir):
        api.current_logger().debug('Mounts directory is a mounted disk image - Unmounting.')
        try:
            run(['/bin/umount', '-fl', mounts_dir])
            api.current_logger().debug('Unmounted mounted disk image.')
        except (OSError, CalledProcessError) as e:
            api.current_logger().warning('Failed to umount %s - message: %s', mounts_dir, str(e))
    api.current_logger().debug('Recursively removing scratch directory %s.', scratch_dir)
    shutil.rmtree(scratch_dir, onerror=utils.report_and_ignore_shutil_rmtree_error)
    api.current_logger().debug('Recursively removed scratch directory %s.', scratch_dir)


def _create_mount_disk_image(disk_images_directory, path):
    """
    Creates the mount disk image, for cases when we hit XFS with ftype=0
    """
    diskimage_path = os.path.join(disk_images_directory, _mount_name(path))
    disk_size = _overlay_disk_size()

    api.current_logger().debug('Attempting to create disk image with size %d MiB at %s', disk_size, diskimage_path)
    utils.call_with_failure_hint(
        cmd=['/bin/dd', 'if=/dev/zero', 'of={}'.format(diskimage_path), 'bs=1M', 'count={}'.format(disk_size)],
        hint='Please ensure that there is enough diskspace in {} at least {} MiB are needed'.format(
            diskimage_path, disk_size)
    )

    api.current_logger().debug('Creating ext4 filesystem in disk image at %s', diskimage_path)
    try:
        utils.call_with_oserror_handled(cmd=['/sbin/mkfs.ext4', '-F', diskimage_path])
    except CalledProcessError as e:
        api.current_logger().error('Failed to create ext4 filesystem %s', exc_info=True)
        raise StopActorExecutionError(
            message=str(e)
        )

    return diskimage_path


def _create_diskimages_dir(scratch_dir, diskimages_dir):
    """
    Prepares directories for disk images
    """
    api.current_logger().debug('Creating disk images directory.')
    try:
        utils.makedirs(diskimages_dir)
        api.current_logger().debug('Done creating disk images directory.')
    except OSError:
        api.current_logger().error('Failed to create disk images directory %s', diskimages_dir, exc_info=True)

        # This is an attempt for giving the user a chance to resolve it on their own
        raise StopActorExecutionError(
            message='Failed to prepare environment for package download while creating directories.',
            details={
                'hint': 'Please ensure that {scratch_dir} is empty and modifiable.'.format(scratch_dir=scratch_dir)
            }
        )


def _create_mounts_dir(scratch_dir, mounts_dir):
    """
    Prepares directories for mounts
    """
    api.current_logger().debug('Creating mount directories.')
    try:
        utils.makedirs(mounts_dir)
        api.current_logger().debug('Done creating mount directories.')
    except OSError:
        api.current_logger().error('Failed to create mounting directories %s', mounts_dir, exc_info=True)

        # This is an attempt for giving the user a chance to resolve it on their own
        raise StopActorExecutionError(
            message='Failed to prepare environment for package download while creating directories.',
            details={
                'hint': 'Please ensure that {scratch_dir} is empty and modifiable.'.format(scratch_dir=scratch_dir)
            }
        )


@contextlib.contextmanager
def _mount_dnf_cache(overlay_target):
    """
    Convenience context manager to ensure bind mounted /var/cache/dnf and removal of the mount.
    """
    with mounting.BindMount(
            source='/var/cache/dnf',
            target=os.path.join(overlay_target, 'var', 'cache', 'dnf')) as cache_mount:
        yield cache_mount


@contextlib.contextmanager
def create_source_overlay(mounts_dir, scratch_dir, xfs_info, storage_info, mount_target=None):
    """
    Context manager that prepares the source system overlay and yields the mount.
    """
    api.current_logger().debug('Creating source overlay in {scratch_dir} with mounts in {mounts_dir}'.format(
        scratch_dir=scratch_dir, mounts_dir=mounts_dir))
    try:
        _create_mounts_dir(scratch_dir, mounts_dir)
        mounts = _prepare_required_mounts(scratch_dir, mounts_dir, _get_mountpoints(storage_info), xfs_info)
        with mounts.pop('/') as root_mount:
            with mounting.OverlayMount(name='system_overlay', source='/', workdir=root_mount.target) as root_overlay:
                if mount_target:
                    target = mounting.BindMount(source=root_overlay.target, target=mount_target)
                else:
                    target = mounting.NullMount(target=root_overlay.target)
                with target:
                    with _build_overlay_mount(root_overlay, mounts) as overlay:
                        with _mount_dnf_cache(overlay.target):
                            yield overlay
    except Exception:
        cleanup_scratch(scratch_dir, mounts_dir)
        raise



import os

from leapp.actors import Actor
from leapp.models import RootDirectory, RootSubdirectory
from leapp.tags import IPUWorkflowTag, FactsPhaseTag


class RootScanner(Actor):
    """
    Scan the system root directory and produce a message containing
    information about its subdirectories.
    """

    name = 'root_scanner'
    consumes = ()
    produces = (RootDirectory,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        subdirs = []
        for subdir in os.listdir('/'):
            path = os.path.join('/', subdir)
            if os.path.islink(path):
                subdirs.append(RootSubdirectory(name=subdir, target=os.readlink(path)))
            else:
                subdirs.append(RootSubdirectory(name=subdir))
        self.produce(RootDirectory(items=subdirs))



from leapp.actors import Actor
from leapp.libraries.actor import initrdinclude
from leapp.models import InitrdIncludes, InstalledTargetKernelVersion
from leapp.tags import IPUWorkflowTag, FinalizationPhaseTag


class InitrdInclude(Actor):
    """
    Regenerate RHEL-8 initrd and include files produced by other actors
    """

    name = 'initrdinclude'
    consumes = (InitrdIncludes, InstalledTargetKernelVersion)
    produces = ()
    tags = (FinalizationPhaseTag, IPUWorkflowTag)

    def process(self):
        initrdinclude.process()



import pytest

from leapp.libraries.actor import checknonmountboots390


class CheckNonMountBootS390ReportCreated(Exception):
    pass


@pytest.mark.parametrize(
    'matches_arch,ismount,should_report', (
        (True, True, False),
        (True, False, True),
        (False, True, False),
        (False, False, False),
    )
)
def test_checknonmountboots390_perform_check(monkeypatch, matches_arch, ismount, should_report):
    def _create_report(data):
        raise CheckNonMountBootS390ReportCreated()

    monkeypatch.setattr(checknonmountboots390.architecture, 'matches_architecture', lambda x: matches_arch)
    monkeypatch.setattr(checknonmountboots390.os.path, 'ismount', lambda x: ismount)
    monkeypatch.setattr(checknonmountboots390.reporting, 'create_report', _create_report)

    if should_report:
        with pytest.raises(CheckNonMountBootS390ReportCreated):
            checknonmountboots390.perform_check()
    else:
        checknonmountboots390.perform_check()



from os import symlink, path, access, X_OK

import pytest

from leapp.libraries.actor import workaround
from leapp.libraries.common.utils import makedirs


def fake_symlink(basedir):
    def impl(source, target):
        source_path = str(basedir.join(*source.lstrip('/').split('/')))
        makedirs(source_path)
        symlink(source_path, target)
    return impl


def test_apply_python3_workaround(monkeypatch, tmpdir):
    leapp_home = tmpdir.mkdir('tmp_leapp_py3')
    monkeypatch.setattr(workaround.os, 'symlink', fake_symlink(tmpdir.mkdir('lib')))
    monkeypatch.setattr(workaround, 'LEAPP_HOME', str(leapp_home))

    # Ensure double invocation doesn't cause a problem
    workaround.apply_python3_workaround()
    workaround.apply_python3_workaround()

    # Ensure creation of all required elements
    assert path.islink(str(leapp_home.join('leapp')))
    assert path.isfile(str(leapp_home.join('leapp3')))
    assert access(str(leapp_home.join('leapp3')), X_OK)

    assert str(leapp_home) in leapp_home.join('leapp3').read_text('utf-8')



import itertools
import logging
from functools import partial

import pytest

from leapp import reporting
from leapp.libraries.actor.checkpcidrivers import (
    checkpcidrivers_main,
    create_dict_lookup,
    render_report,
)
from leapp.libraries.common.testutils import (
    CurrentActorMocked,
    create_report_mocked,
    produce_mocked,
)
from leapp.libraries.stdlib import api
from leapp.models import (
    Model,
    PCIDevice,
    PCIDevices,
    RestrictedPCIDevice,
    RestrictedPCIDevices,
    fields,
)
from leapp.topics.systeminfo import SystemInfoTopic

logger = logging.getLogger(__name__)

pci_devices = PCIDevices(
    devices=[
        PCIDevice(
            slot="00:00.0",
            driver="",
            dev_cls="Host bridge",
            vendor="Intel Corporation",
            name="440FX - 82441FX PMC [Natoma]",
            subsystem_vendor="Red Hat, Inc.",
            subsystem_name="Qemu virtual machine",
            pci_id="",
            rev="02",
        ),
        PCIDevice(
            slot="00:01.0",
            dev_cls="ISA bridge",
            vendor="Intel Corporation",
            name="82371SB PIIX3 ISA [Natoma/Triton II]",
            subsystem_vendor="Red Hat, Inc.",
            pci_id="15b560:0739",
            subsystem_name="Qemu virtual machine",
        ),
        PCIDevice(
            slot="00:01.1",
            dev_cls="IDE interface",
            vendor="Intel Corporation",
            name="82371SB PIIX3 IDE [Natoma/Triton II]",
            subsystem_vendor="Red Hat, Inc.",
            subsystem_name="Qemu virtual machine",
            pci_id="15b560:0739",
            progif="80",
        ),
    ]
)

restricted_pci_devices = RestrictedPCIDevices(
    driver_names=[
        RestrictedPCIDevice(
            pci_id="nan",
            driver_name="3w-9xxx",
            device_name="3w-9xxx",
            available_rhel7=1,
            supported_rhel7=1,
            available_rhel8=0,
            supported_rhel8=0,
            available_rhel9=0,
            supported_rhel9=0,
            comment="nan",
        ),
        RestrictedPCIDevice(
            pci_id="nan",
            driver_name="3w-sas",
            device_name="3w-sas",
            available_rhel7=1,
            supported_rhel7=1,
            available_rhel8=1,
            supported_rhel8=0,
            available_rhel9=0,
            supported_rhel9=0,
            comment="nan",
        ),
        RestrictedPCIDevice(
            pci_id="nan",
            driver_name="acard-ahci",
            device_name="acard-ahci",
            available_rhel7=1,
            supported_rhel7=1,
            available_rhel8=1,
            supported_rhel8=1,
            available_rhel9=0,
            supported_rhel9=0,
            comment="nan",
        ),
    ],
    pci_ids=[
        RestrictedPCIDevice(
            pci_id="0x1000:0x0060",
            driver_name="megaraid_sas",
            device_name="SAS1078R",
            available_rhel7=1,
            supported_rhel7=1,
            available_rhel8=0,
            supported_rhel8=0,
            available_rhel9=0,
            supported_rhel9=0,
            comment="nan",
        ),
        RestrictedPCIDevice(
            pci_id="0x1000:0x0064",
            driver_name="mpt2sas",
            device_name="SAS2116_1",
            available_rhel7=1,
            supported_rhel7=1,
            available_rhel8=1,
            supported_rhel8=0,
            available_rhel9=0,
            supported_rhel9=0,
            comment="nan",
        ),
        RestrictedPCIDevice(
            pci_id="0x1000:0x0065",
            driver_name="mpt2sas",
            device_name="SAS2116_2",
            available_rhel7=1,
            supported_rhel7=1,
            available_rhel8=1,
            supported_rhel8=1,
            available_rhel9=0,
            supported_rhel9=0,
            comment="nan",
        ),
    ],
)


@pytest.mark.parametrize(
    (
        "driver_name",
        "pci_id",
        "inhibit_upgrade",
        "num_msgs",
    ),
    [
        (
            "good driver name",
            "good pci id",
            None,
            0,
        ),
        (
            "3w-9xxx",
            "good pci id",
            True,
            1,
        ),
        (
            "3w-sas",
            "good pci id",
            None,
            1,
        ),
        (
            "acard-ahci",
            "good pci id",
            None,
            0,
        ),
        (
            "good driver name",
            "0x1000:0x0060",
            True,
            1,
        ),
        (
            "good driver name",
            "0x1000:0x0064",
            None,
            1,
        ),
        (
            "good driver name",
            "0x1000:0x0065",
            None,
            0,
        ),
    ],
)
def test_basic_checkpcidrivers(
    monkeypatch,
    driver_name,
    pci_id,
    inhibit_upgrade,
    num_msgs,
):
    """Check the main actor function."""
    pci_devices.devices[0].driver = driver_name
    pci_devices.devices[0].pci_id = pci_id
    monkeypatch.setattr(
        api,
        "current_actor",
        CurrentActorMocked(msgs=[pci_devices, restricted_pci_devices]),
    )
    monkeypatch.setattr(
        reporting,
        "create_report",
        create_report_mocked(),
    )
    monkeypatch.setattr(
        CurrentActorMocked,
        "produce",
        produce_mocked(),
    )
    checkpcidrivers_main()
    if inhibit_upgrade:
        assert len(CurrentActorMocked.produce.model_instances) == 1
        assert CurrentActorMocked.produce.model_instances[0].report[
            "flags"
        ] == ["inhibitor"]
    if num_msgs and not inhibit_upgrade:
        assert (
            "flags" not in CurrentActorMocked.produce.model_instances[0].report
        )
    assert len(CurrentActorMocked.produce.model_instances) == num_msgs


RestrictedPCIDevicePrefilled = partial(
    RestrictedPCIDevice,
    pci_id="",
    driver_name="",
    device_name="",
    comment="",
    available_rhel7=1,
    available_rhel9=0,
    supported_rhel7=1,
    supported_rhel9=0,
)


def get_test_case_for_testing_report(
    devices, available_rhel8, supported_rhel8
):
    return devices, {
        d: RestrictedPCIDevicePrefilled(
            available_rhel8=available_rhel8,
            supported_rhel8=supported_rhel8,
        )
        for d in devices
    }


@pytest.mark.parametrize(
    (
        "restricted_driver_names_on_host",
        "restricted_devices_drivers",
        "restricted_pci_ids_on_host",
        "restricted_devices_pcis",
        "inhibit_upgrade",
        "exp_line_length",
    ),
    [
        get_test_case_for_testing_report(("d1", "d2"), 0, 0)
        + get_test_case_for_testing_report(("p1", "p2"), 0, 0)
        + (True, 15),
        get_test_case_for_testing_report(("d1", "d2"), 1, 0)
        + get_test_case_for_testing_report(("p1", "p2"), 1, 0)
        + (True, 8),
        get_test_case_for_testing_report(("d1", "d2"), 0, 1)
        + get_test_case_for_testing_report(("p1", "p2"), 0, 1)
        + (True, 8),
        get_test_case_for_testing_report(("d1", "d2"), 1, 0)
        + get_test_case_for_testing_report(("p1", "p2"), 0, 1)
        + (True, 9),
    ],
)
def test_render_report(
    monkeypatch,
    restricted_driver_names_on_host,
    restricted_pci_ids_on_host,
    restricted_devices_drivers,
    restricted_devices_pcis,
    inhibit_upgrade,
    exp_line_length,
    caplog,
):
    """Check report appearance."""
    monkeypatch.setattr(
        reporting,
        "create_report",
        create_report_mocked(),
    )
    reporting.create_report(
        render_report(
            restricted_driver_names_on_host=restricted_driver_names_on_host,
            restricted_pci_ids_on_host=restricted_pci_ids_on_host,
            restricted_devices_drivers=restricted_devices_drivers,
            restricted_devices_pcis=restricted_devices_pcis,
            inhibit_upgrade=inhibit_upgrade,
        )
    )
    logger.info(reporting.create_report.report_fields["summary"])
    assert sum(len(m.split("\n")) for m in caplog.messages) == exp_line_length


class M(Model):
    """A model just for the purpose of further test."""

    topic = SystemInfoTopic

    a = fields.Integer(default=0)
    b = fields.Integer(default=0)


@pytest.mark.parametrize(
    ("driver_names", "key", "exp", "exception"),
    [
        # Normal case
        (
            [M(a=1, b=2), M(a=1, b=3)],
            "b",
            {2: M(a=1, b=2), 3: M(a=1, b=3)},
            None,
        ),
        # Not existing attribute
        (
            [M(a=1, b=2), M(a=1, b=3)],
            "c",
            None,
            AttributeError,
        ),
        # Not existing attribute, empty Ms
        (
            [M(), M()],
            "c",
            None,
            AttributeError,
        ),
        # Duplicated attribute values
        (
            [M(a=1, b=2), M(a=1, b=2)],
            "b",
            None,
            ValueError,
        ),
    ],
)
def test_create_dict_lookup_fn(driver_names, key, exp, exception):
    if exception:
        with pytest.raises(exception):
            create_dict_lookup(driver_names, key)
    else:
        assert create_dict_lookup(driver_names, key) == exp


def test_checkpcidrivers_no_restricted_devs(current_actor_context):
    """Test if actor works well when no messages provided."""
    current_actor_context.run()
    assert current_actor_context._messaging.errors()



from leapp.libraries.common.config import mock_configs
from leapp.models import GrubDevice, UpdateGrub, FirmwareFacts
from leapp.reporting import Report


def test_actor_update_grub(current_actor_context):
    current_actor_context.feed(FirmwareFacts(firmware='bios'))
    current_actor_context.feed(GrubDevice(grub_device='/dev/vda'))
    current_actor_context.run(config_model=mock_configs.CONFIG)
    assert current_actor_context.consume(Report)
    assert current_actor_context.consume(UpdateGrub)
    assert current_actor_context.consume(UpdateGrub)[0].grub_device == '/dev/vda'


def test_actor_no_grub_device(current_actor_context):
    current_actor_context.feed(FirmwareFacts(firmware='bios'))
    current_actor_context.run(config_model=mock_configs.CONFIG)
    assert current_actor_context.consume(Report)
    assert not current_actor_context.consume(UpdateGrub)


def test_actor_with_efi(current_actor_context):
    current_actor_context.feed(FirmwareFacts(firmware='efi'))
    current_actor_context.run(config_model=mock_configs.CONFIG)
    assert not current_actor_context.consume(Report)
    assert not current_actor_context.consume(UpdateGrub)


def test_s390x(current_actor_context):
    current_actor_context.feed(FirmwareFacts(firmware='bios'))
    current_actor_context.feed(GrubDevice(grub_device='/dev/vda'))
    current_actor_context.run(config_model=mock_configs.CONFIG_S390X)
    assert not current_actor_context.consume(Report)
    assert not current_actor_context.consume(UpdateGrub)



import warnings

from leapp.exceptions import StopActorExecutionError

no_yum = False
no_yum_warning_msg = "package `yum` is unavailable"
try:
    import yum
except ImportError:
    no_yum = True
    warnings.warn(no_yum_warning_msg, ImportWarning)


def get_package_repository_data():
    """ Return dictionary mapping package name with repository from which it was installed.
    Note:
        There's no yum module for py3. The dnf module could have been used
        instead but there's a bug in dnf preventing us to do so:
        https://bugzilla.redhat.com/show_bug.cgi?id=1789840
    """
    if no_yum:
        raise StopActorExecutionError(message=no_yum_warning_msg)
    yum_base = yum.YumBase()
    pkg_repos = {}

    try:
        for pkg in yum_base.doPackageLists().installed:
            pkg_repos[pkg.name] = pkg.ui_from_repo.lstrip('@')
    except ValueError as e:
        if 'locale' not in str(e):  # reraise if error is not related to locales
            raise e
        raise StopActorExecutionError(
            message='Failed to get installed RPM packages because of an invalid locale',
            details={
                'hint': 'Please run leapp with a valid locale. ' +
                        'You can get a list of installed locales by running `locale -a`.'
            })

    return pkg_repos



import os
from collections import namedtuple

import pytest

from leapp import models
from leapp.exceptions import StopActorExecution, StopActorExecutionError
from leapp.libraries.actor import userspacegen
from leapp.libraries.common import overlaygen, repofileutils, rhsm
from leapp.libraries.common.config import architecture
from leapp.libraries.common.testutils import CurrentActorMocked, logger_mocked, produce_mocked

CUR_DIR = os.path.dirname(os.path.abspath(__file__))
_CERTS_PATH = os.path.join(CUR_DIR, '../../../files', userspacegen.PROD_CERTS_FOLDER)
_DEFAULT_CERT_PATH = os.path.join(_CERTS_PATH, '8.1', '479.pem')


@pytest.fixture
def adjust_cwd():
    previous_cwd = os.getcwd()
    os.chdir(os.path.join(CUR_DIR, "../"))
    yield
    os.chdir(previous_cwd)


class MockedMountingBase(object):
    def __init__(self, **dummy_kwargs):
        self.called_copytree_from = []

    def copytree_from(self, src, dst):
        self.called_copytree_from.append((src, dst))

    def __call__(self, **dummy_kwarg):
        yield self

    def call(self, *args, **kwargs):
        return {'stdout': ''}

    def nspawn(self):
        return self

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        pass


@pytest.mark.parametrize('result,dst_ver,arch,prod_type', [
    (os.path.join(_CERTS_PATH, '8.1', '479.pem'), '8.1', architecture.ARCH_X86_64, 'ga'),
    (os.path.join(_CERTS_PATH, '8.1', '419.pem'), '8.1', architecture.ARCH_ARM64, 'ga'),
    (os.path.join(_CERTS_PATH, '8.1', '279.pem'), '8.1', architecture.ARCH_PPC64LE, 'ga'),
    (os.path.join(_CERTS_PATH, '8.2', '479.pem'), '8.2', architecture.ARCH_X86_64, 'ga'),
    (os.path.join(_CERTS_PATH, '8.4', '230.pem'), '8.4', architecture.ARCH_X86_64, 'htb'),
    (os.path.join(_CERTS_PATH, '8.4', '486.pem'), '8.4', architecture.ARCH_X86_64, 'beta'),
    (os.path.join(_CERTS_PATH, '8.2', '72.pem'), '8.2', architecture.ARCH_S390X, 'ga'),
    (os.path.join(_CERTS_PATH, '8.4', '232.pem'), '8.4', architecture.ARCH_S390X, 'htb'),
    (os.path.join(_CERTS_PATH, '8.4', '433.pem'), '8.4', architecture.ARCH_S390X, 'beta'),
])
def test_get_product_certificate_path(monkeypatch, adjust_cwd, result, dst_ver, arch, prod_type):
    envars = {'LEAPP_DEVEL_TARGET_PRODUCT_TYPE': prod_type}
    curr_actor_mocked = CurrentActorMocked(dst_ver=dst_ver, arch=arch, envars=envars)
    monkeypatch.setattr(userspacegen.api, 'current_actor', curr_actor_mocked)
    assert userspacegen._get_product_certificate_path() in result


_PACKAGES_MSGS = [
    models.RequiredTargetUserspacePackages(),
    models.RequiredTargetUserspacePackages(packages=['pkgA']),
    models.RequiredTargetUserspacePackages(packages=['pkgB', 'pkgsC']),
    models.RequiredTargetUserspacePackages(packages=['pkgD']),
]
_RHSMINFO_MSG = models.RHSMInfo(attached_skus=['testing-sku'])
_RHUIINFO_MSG = models.RHUIInfo(provider='aws')
_XFS_MSG = models.XFSPresence()
_STORAGEINFO_MSG = models.StorageInfo()
_CTRF_MSGS = [
    models.CustomTargetRepositoryFile(file='rfileA'),
    models.CustomTargetRepositoryFile(file='rfileB'),
]
_SAEE = StopActorExecutionError
_SAE = StopActorExecution


class MockedConsume(object):
    def __init__(self, *args):
        self._msgs = []
        for arg in args:
            if not arg:
                continue
            if isinstance(arg, list):
                self._msgs.extend(arg)
            else:
                self._msgs.append(arg)

    def __call__(self, model):
        return iter([msg for msg in self._msgs if isinstance(msg, model)])


testInData = namedtuple(
    'TestInData', ['pkg_msgs', 'rhsm_info', 'rhui_info', 'xfs', 'storage', 'custom_repofiles']
)


@pytest.mark.parametrize('raised,no_rhsm,testdata', [
    # valid cases with RHSM
    (None, '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    (None, '0', testInData(_PACKAGES_MSGS[0], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    (None, '0', testInData([], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    (None, '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, None, _STORAGEINFO_MSG, None)),
    (None, '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, _CTRF_MSGS)),
    (None, '0', testInData(_PACKAGES_MSGS[0], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, _CTRF_MSGS)),
    (None, '0', testInData([], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, _CTRF_MSGS)),
    (None, '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, None, _STORAGEINFO_MSG, _CTRF_MSGS)),

    # valid cases without RHSM (== skip_rhsm)
    (None, '1', testInData(_PACKAGES_MSGS, None, _RHUIINFO_MSG, _XFS_MSG, _STORAGEINFO_MSG, None)),
    (None, '1', testInData(_PACKAGES_MSGS, None, _RHUIINFO_MSG, None, _STORAGEINFO_MSG, None)),
    (None, '1', testInData([], None, _RHUIINFO_MSG, _XFS_MSG, _STORAGEINFO_MSG, None)),
    (None, '1', testInData([], None, _RHUIINFO_MSG, None, _STORAGEINFO_MSG, None)),
    (None, '1', testInData(_PACKAGES_MSGS, None, _RHUIINFO_MSG, _XFS_MSG, _STORAGEINFO_MSG, _CTRF_MSGS)),
    (None, '1', testInData(_PACKAGES_MSGS, None, _RHUIINFO_MSG, None, _STORAGEINFO_MSG, _CTRF_MSGS)),
    (None, '1', testInData([], None, _RHUIINFO_MSG, _XFS_MSG, _STORAGEINFO_MSG, _CTRF_MSGS)),
    (None, '1', testInData([], None, _RHUIINFO_MSG, None, _STORAGEINFO_MSG, _CTRF_MSGS)),

    # no-rhsm but RHSMInfo defined (should be _RHSMINFO_MSG)
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS[0], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG,
                                             None)),
    ((_SAEE, 'RHSM is not'), '1', testInData([], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, None, _STORAGEINFO_MSG, None)),
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG,
                                             _CTRF_MSGS)),
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS[0], _RHSMINFO_MSG, None, _XFS_MSG,
                                             _STORAGEINFO_MSG, _CTRF_MSGS)),
    ((_SAEE, 'RHSM is not'), '1', testInData([], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, _CTRF_MSGS)),
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, None, _STORAGEINFO_MSG, _CTRF_MSGS)),

    # missing RHSMInfo but it should exist
    # NOTE: should be this Error?!
    ((_SAE, 'RHSM information'), '0', testInData(_PACKAGES_MSGS, None, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    ((_SAE, 'RHSM information'), '0', testInData(_PACKAGES_MSGS, None, None, None, _STORAGEINFO_MSG, None)),
    ((_SAE, 'RHSM information'), '0', testInData([], None, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    ((_SAE, 'RHSM information'), '0', testInData([], None, None, None, _STORAGEINFO_MSG, None)),

    # in the end, error when StorageInfo is missing
    ((_SAEE, 'No storage'), '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, _XFS_MSG, None, None)),
    ((_SAEE, 'No storage'), '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, None, None, None)),
    ((_SAEE, 'No storage'), '0', testInData([], _RHSMINFO_MSG, None, _XFS_MSG, None, None)),
    ((_SAEE, 'No storage'), '0', testInData([], _RHSMINFO_MSG, None, None, None, None)),
    ((_SAEE, 'No storage'), '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, _XFS_MSG, None, _CTRF_MSGS)),
    ((_SAEE, 'No storage'), '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, None, None, _CTRF_MSGS)),
    ((_SAEE, 'No storage'), '0', testInData([], _RHSMINFO_MSG, None, _XFS_MSG, None, _CTRF_MSGS)),
    ((_SAEE, 'No storage'), '0', testInData([], _RHSMINFO_MSG, None, None, None, _CTRF_MSGS)),
])
def test_consume_data(monkeypatch, raised, no_rhsm, testdata):
    # do not write never into testdata inside the test !!
    xfs = testdata.xfs
    custom_repofiles = testdata.custom_repofiles
    _exp_pkgs = {'dnf'}
    if isinstance(testdata.pkg_msgs, list):
        for msg in testdata.pkg_msgs:
            _exp_pkgs.update(msg.packages)
    else:
        _exp_pkgs.update(testdata.pkg_msgs.packages)
    mocked_consume = MockedConsume(testdata.pkg_msgs,
                                   testdata.rhsm_info,
                                   testdata.rhui_info,
                                   xfs,
                                   testdata.storage,
                                   custom_repofiles)
    monkeypatch.setattr(userspacegen.api, 'consume', mocked_consume)
    monkeypatch.setattr(userspacegen.api, 'current_logger', logger_mocked())
    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked(envars={'LEAPP_NO_RHSM': no_rhsm}))
    if not xfs:
        xfs = models.XFSPresence()
    if not custom_repofiles:
        custom_repofiles = []
    if not raised:
        result = userspacegen._InputData()
        assert result.packages == _exp_pkgs
        assert result.rhsm_info == testdata.rhsm_info
        assert result.rhui_info == testdata.rhui_info
        assert result.xfs_info == xfs
        assert result.storage_info == testdata.storage
        assert result.custom_repofiles == custom_repofiles
        assert not userspacegen.api.current_logger.warnmsg
        assert not userspacegen.api.current_logger.errmsg
    else:
        with pytest.raises(raised[0]) as err:
            userspacegen._InputData()
        if isinstance(err.value, StopActorExecutionError):
            assert raised[1] in err.value.message
        else:
            assert userspacegen.api.current_logger.warnmsg
            assert any([raised[1] in x for x in userspacegen.api.current_logger.warnmsg])


@pytest.mark.skip(reason="Currently not implemented in the actor. It's TODO.")
def test_gather_target_repositories(monkeypatch):
    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked())
    # The available RHSM repos
    monkeypatch.setattr(rhsm, 'get_available_repo_ids', lambda x: ['repoidX', 'repoidY', 'repoidZ'])
    monkeypatch.setattr(rhsm, 'skip_rhsm', lambda: False)
    # The required RHEL repos based on the repo mapping and PES data + custom repos required by third party actors
    monkeypatch.setattr(userspacegen.api, 'consume', lambda x: iter([models.TargetRepositories(
        rhel_repos=[models.RHELTargetRepository(repoid='repoidX'),
                    models.RHELTargetRepository(repoid='repoidY')],
        custom_repos=[models.CustomTargetRepository(repoid='repoidCustom')])]))

    target_repoids = userspacegen.gather_target_repositories(None, None)

    assert target_repoids == ['repoidX', 'repoidY', 'repoidCustom']


def test_gather_target_repositories_none_available(monkeypatch):

    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(rhsm, 'get_available_repo_ids', lambda x: [])
    monkeypatch.setattr(rhsm, 'skip_rhsm', lambda: False)
    with pytest.raises(StopActorExecutionError) as err:
        userspacegen.gather_target_repositories(None, None)
    assert "Cannot find required basic RHEL 8 repositories" in str(err)


def test_gather_target_repositories_rhui(monkeypatch):

    indata = testInData(
        _PACKAGES_MSGS, _RHSMINFO_MSG, _RHUIINFO_MSG, _XFS_MSG, _STORAGEINFO_MSG, None
    )

    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(userspacegen, '_get_all_available_repoids', lambda x: [])
    monkeypatch.setattr(
        userspacegen, '_get_rh_available_repoids', lambda x, y: ['rhui-1', 'rhui-2', 'rhui-3']
    )
    monkeypatch.setattr(rhsm, 'skip_rhsm', lambda: True)
    monkeypatch.setattr(
        userspacegen.api, 'consume', lambda x: iter(
            [models.TargetRepositories(
                rhel_repos=[
                    models.RHELTargetRepository(repoid='rhui-1'),
                    models.RHELTargetRepository(repoid='rhui-2')
                ]
            )
            ])
    )
    target_repoids = userspacegen.gather_target_repositories(None, indata)
    assert target_repoids == set(['rhui-1', 'rhui-2'])


@pytest.mark.skip(reason="Currently not implemented in the actor. It's TODO.")
def test_gather_target_repositories_required_not_available(monkeypatch):
    # If the repos that Leapp identifies as required for the upgrade (based on the repo mapping and PES data) are not
    # available, an exception shall be raised

    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked())
    # The available RHSM repos
    monkeypatch.setattr(rhsm, 'get_available_repo_ids', lambda x: ['repoidA', 'repoidB', 'repoidC'])
    monkeypatch.setattr(rhsm, 'skip_rhsm', lambda: False)
    # The required RHEL repos based on the repo mapping and PES data + custom repos required by third party actors
    monkeypatch.setattr(userspacegen.api, 'consume', lambda x: iter([models.TargetRepositories(
        rhel_repos=[models.RHELTargetRepository(repoid='repoidX'),
                    models.RHELTargetRepository(repoid='repoidY')],
        custom_repos=[models.CustomTargetRepository(repoid='repoidCustom')])]))

    with pytest.raises(StopActorExecutionError) as err:
        userspacegen.gather_target_repositories(None)
    assert "Cannot find required basic RHEL 8 repositories" in str(err)


def mocked_consume_data():
    packages = {'dnf', 'pkgA', 'pkgB'}
    rhsm_info = _RHSMINFO_MSG
    rhui_info = _RHUIINFO_MSG
    xfs_info = models.XFSPresence()
    storage_info = models.StorageInfo()
    custom_repofiles = []
    fields = ['packages', 'rhsm_info', 'rhui_info', 'xfs_info', 'storage_info', 'custom_repofiles']

    return namedtuple('TestInData', fields)(
                packages, rhsm_info, rhui_info, xfs_info, storage_info, custom_repofiles
    )


# TODO: come up with additional tests for the main function
def test_perform_ok(monkeypatch):
    repoids = ['repoidX', 'repoidY']
    monkeypatch.setattr(userspacegen, '_InputData', mocked_consume_data)
    monkeypatch.setattr(userspacegen, '_get_product_certificate_path', lambda: _DEFAULT_CERT_PATH)
    monkeypatch.setattr(overlaygen, 'create_source_overlay', MockedMountingBase)
    monkeypatch.setattr(userspacegen, '_gather_target_repositories', lambda *x: repoids)
    monkeypatch.setattr(userspacegen, '_create_target_userspace', lambda *x: None)
    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(userspacegen.api, 'produce', produce_mocked())
    monkeypatch.setattr(repofileutils, 'get_repodirs', lambda: ['/etc/yum.repos.d'])
    userspacegen.perform()
    msg_target_repos = models.UsedTargetRepositories(
        repos=[models.UsedTargetRepository(repoid=repo) for repo in repoids])
    assert userspacegen.api.produce.called == 3
    assert isinstance(userspacegen.api.produce.model_instances[0], models.TMPTargetRepositoriesFacts)
    assert userspacegen.api.produce.model_instances[1] == msg_target_repos
    # this one is full of contants, so it's safe to check just the instance
    assert isinstance(userspacegen.api.produce.model_instances[2], models.TargetUserSpaceInfo)



import os

from leapp.libraries.actor import scancpu
from leapp.libraries.common import testutils
from leapp.libraries.stdlib import api
from leapp.models import CPUInfo


CUR_DIR = os.path.dirname(os.path.abspath(__file__))


class mocked_get_cpuinfo(object):
    def __init__(self, filename):
        self.filename = filename

    def __call__(self):
        """
        Return lines of the self.filename test file located in the files directory.

        Those files contain /proc/cpuinfo content from several machines.
        """
        with open(os.path.join(CUR_DIR, 'files', self.filename), 'r') as fp:
            return fp.readlines()


def test_machine_type(monkeypatch):
    # cpuinfo doesn't contain a machine field
    mocked_cpuinfo = mocked_get_cpuinfo('cpuinfo_x86_64')
    monkeypatch.setattr(scancpu, '_get_cpuinfo', mocked_cpuinfo)
    monkeypatch.setattr(api, 'produce', testutils.produce_mocked())
    scancpu.process()
    assert api.produce.called == 1
    assert CPUInfo() == api.produce.model_instances[0]

    # cpuinfo contains a machine field
    api.produce.called = 0
    api.produce.model_instances = []
    mocked_cpuinfo.filename = 'cpuinfo_s390x'
    scancpu.process()
    assert api.produce.called == 1
    assert CPUInfo(machine_type=2827) == api.produce.model_instances[0]



from leapp.actors import Actor
from leapp.exceptions import StopActorExecutionError
from leapp.models import Report, KernelCmdline
from leapp.tags import IPUWorkflowTag, ChecksPhaseTag
from leapp import reporting


class CheckFips(Actor):
    """
    Inhibit upgrade if FIPS is detected as enabled.
    """

    name = 'check_fips'
    consumes = (KernelCmdline,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        cmdline = next(self.consume(KernelCmdline), None)
        if not cmdline:
            raise StopActorExecutionError('Cannot check FIPS state due to missing command line parameters',
                                          details={'Problem': 'Did not receive a message with kernel command '
                                                              'line parameters (KernelCmdline)'})
        for parameter in cmdline.parameters:
            if parameter.key == 'fips' and parameter.value == '1':
                title = 'Cannot upgrade a system with FIPS mode enabled'
                summary = 'Leapp has detected that FIPS is enabled on this system. ' \
                          'In-place upgrade of systems in FIPS mode is currently unsupported.'
                reporting.create_report([
                    reporting.Title(title),
                    reporting.Summary(summary),
                    reporting.Severity(reporting.Severity.HIGH),
                    reporting.Tags([reporting.Tags.SECURITY]),
                    reporting.Flags([reporting.Flags.INHIBITOR])
                ])



from leapp.snactor.fixture import current_actor_context
from leapp.models import IpaInfo
from leapp.reporting import Report


def mock_ipa_info(client, server_pkg, server_configured):
    return IpaInfo(
        has_client_package=client,
        is_client_configured=client,
        has_server_package=server_pkg,
        is_server_configured=server_configured,
    )


def test_inhibit_ipa_configured(current_actor_context):
    current_actor_context.feed(mock_ipa_info(True, True, True))
    current_actor_context.run()
    reports = current_actor_context.consume(Report)

    assert len(reports) == 1
    fields = reports[0].report
    assert "inhibitor" in fields["flags"]
    assert "ipa-server" in fields["title"]


def test_warn_server_pkg(current_actor_context):
    current_actor_context.feed(mock_ipa_info(True, True, False))
    current_actor_context.run()
    reports = current_actor_context.consume(Report)

    assert len(reports) == 1
    fields = reports[0].report
    assert "flags" not in fields
    assert "ipa-server" in fields["title"]


def test_client_only(current_actor_context):
    current_actor_context.feed(mock_ipa_info(True, False, False))
    current_actor_context.run()
    reports = current_actor_context.consume(Report)

    assert not reports


def test_no_ipa(current_actor_context):
    current_actor_context.feed(mock_ipa_info(False, False, False))
    current_actor_context.run()
    reports = current_actor_context.consume(Report)

    assert not reports



import os

import pytest

from leapp.libraries.stdlib import api, run, CalledProcessError
from leapp.models import SELinuxModule, SELinuxModules, SELinuxCustom
from leapp.reporting import Report
from leapp.snactor.fixture import current_actor_context


TEST_MODULES = [
    ["400", "mock1"],
    ["99", "mock1"],
    ["300", "mock1"],
    ["400", "mock2"],
    ["999", "mock3"]
]

SEMANAGE_COMMANDS = [
    ['fcontext', '-t', 'httpd_sys_content_t', '"/web(/.*)?"'],
    ['fcontext', '-t', 'ganesha_var_run_t', '"/ganesha(/.*)?"'],
    ['port', '-t', 'http_port_t', '-p', 'udp', '81'],
    ['permissive', 'abrt_t']
]

testmoduledir = "tests/mock_modules/"


def _run_cmd(cmd, logmsg="", split=False):
    try:
        return run(cmd, split=split).get("stdout", "")
    except CalledProcessError as e:
        if logmsg:
            api.current_logger().warning("%s: %s", logmsg, str(e.stderr))
    return None


@pytest.fixture(scope="module")
def semodule_lfull_initial():
    yield _run_cmd(["semodule", "-lfull"], logmsg="Error listing SELinux customizations")


@pytest.fixture(scope="module")
def semanage_export_initial():
    yield _run_cmd(["semanage", "export"], logmsg="Error listing SELinux customizations")


@pytest.fixture(scope="function")
def destructive_selinux_env():
    for priority, module in TEST_MODULES:
        tests_dir = os.path.join(os.getenv('PYTEST_CURRENT_TEST').rsplit(os.path.sep, 2)[0], testmoduledir)
        _run_cmd(["semodule", "-X", priority, "-i", os.path.join(tests_dir, module + ".cil")],
                 logmsg="Error installing mock module")

    for command in SEMANAGE_COMMANDS:
        _run_cmd(["semanage", command[0], "-a"] + command[1:], logmsg="Error applying selinux customizations")

    yield

    for priority, module in TEST_MODULES + [["400", "permissive_abrt_t"]]:
        _run_cmd(["semodule", "-X", priority, "-r", module])

    for command in SEMANAGE_COMMANDS:
        _run_cmd(["semanage", command[0], "-d"] + command[1:])


@pytest.mark.skipif(os.getenv("DESTRUCTIVE_TESTING", False) in [False, "0"],
                    reason='Test disabled by default because it would modify the system')
def test_SELinuxPrepare(current_actor_context, semodule_lfull_initial, semanage_export_initial,
                        destructive_selinux_env):
    before_test = []
    for cmd in (["semodule", "-lfull"], ["semanage", "export"]):
        res = _run_cmd(cmd, "Error listing SELinux customizations")
        before_test.append(res)
        # XXX still not sure about logging in tests
        api.current_logger().info("Before test:%s", res)
    # Make sure that initial semodule/semanage commands don't match before tests ones
    assert before_test != [semodule_lfull_initial, semanage_export_initial]

    # XXX FIXME test_modules.reverse() returns None and changes underlying list which should not happen
    # to global vars. Removing that reversing as it is broken anyway
    semodule_list = [SELinuxModule(name=module, priority=int(prio), content="", removed=[])
                     for (prio, module) in TEST_MODULES + [["400", "permissive_abrt_t"]]]

    current_actor_context.feed(SELinuxModules(modules=semodule_list))
    current_actor_context.run()

    # check if all given modules and local customizations where removed
    semodule_res = _run_cmd(["semodule", "-lfull"], "Error listing SELinux modules")
    assert semodule_lfull_initial == semodule_res
    semanage_res = _run_cmd(["semanage", "export"], "Error listing SELinux customizations")
    assert semanage_export_initial == semanage_res



import pytest

from leapp.models import PCIDevice, PCIDevices, KernelModuleParameter
from leapp.reporting import Report


devices_ok = [
    PCIDevice(slot="00:00.0", dev_cls="foo", vendor="bar", name="foobar", driver="i915", pci_id="some id"),
    PCIDevice(slot="00:01.0", dev_cls="foo", vendor="bar", name="foobar", driver="serial", pci_id="some id"),
    PCIDevice(slot="00:02.0", dev_cls="foo", vendor="bar", name="foobar", driver="pcieport", pci_id="some id"),
    PCIDevice(slot="00:03.0", dev_cls="foo", vendor="bar", name="foobar", driver="nvme", pci_id="some id")
]
devices_driverless = [
    PCIDevice(slot="00:04.0", dev_cls="foo", vendor="bar", name="foobar", driver="", pci_id="some id"),
    PCIDevice(slot="00:05.0", dev_cls="foo", vendor="bar", name="foobar", driver="", pci_id="some id")
]
devices_removed = [
    PCIDevice(slot="00:06.0", dev_cls="foo", vendor="bar", name="foobar", driver="floppy", pci_id="some id"),
    PCIDevice(slot="00:07.0", dev_cls="foo", vendor="bar", name="foobar", driver="initio", pci_id="some id"),
    PCIDevice(slot="00:08.0", dev_cls="foo", vendor="bar", name="foobar", driver="pata_acpi", pci_id="some id"),
    PCIDevice(slot="00:09.0", dev_cls="foo", vendor="bar", name="foobar", driver="iwl4965", pci_id="some id")
]


@pytest.mark.parametrize("devices,expected", [
    ([], True),
    (devices_ok, True),
    (devices_driverless, True),
    (devices_ok + devices_driverless, True),
    (devices_removed, False),
    (devices_removed + devices_ok, False),
    (devices_removed + devices_driverless, False),
    (devices_ok + devices_removed + devices_driverless, False)])
def test_drivers(devices, expected, current_actor_context):
    """
    Tests CheckKernelDrivers actor by feeding it mocked PCI devices with their
    respective drivers, if they have one.  Actor should produce a report iff any
    mocked devices from devices_removed are fed to the actor, since their
    drivers are removed in RHEL8 (as per 'files/removed_drivers.txt').
    """
    current_actor_context.feed(PCIDevices(devices=devices))
    current_actor_context.run()
    if expected:
        assert not current_actor_context.consume(Report)
    else:
        assert current_actor_context.consume(Report)



import os
import shutil

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import dnfplugin, mounting
from leapp.libraries.common.config import architecture
from leapp.libraries.stdlib import api
from leapp.models import (BootContent, RequiredUpgradeInitramPackages, TargetUserSpaceInfo, UpgradeDracutModule,
                          UsedTargetRepositories)

INITRAM_GEN_SCRIPT_NAME = 'generate-initram.sh'


def _reinstall_leapp_repository_hint():
    """
    Convenience function for creating a detail for StopActorExecutionError with a hint to reinstall the
    leapp-repository package
    """
    return {
        'hint': 'Try to reinstall the `leapp-repository` package.'
    }


def copy_dracut_modules(context, modules):
    """
    Copies our dracut modules into the target userspace.
    """
    try:
        shutil.rmtree(context.full_path('/dracut'))
    except EnvironmentError:
        pass
    for module in modules:
        try:
            context.copytree_to(module.module_path, os.path.join('/dracut', os.path.basename(module.module_path)))
        except shutil.Error as e:
            api.current_logger().error('Failed to copy dracut module "{name}" from "{source}" to "{target}"'.format(
                name=module.name, source=module.module_path, target=context.full_path('/dracut')), exc_info=True)
            raise StopActorExecutionError(
                message='Failed to install dracut modules required in the initram. Error: {}'.format(str(e))
            )


def _get_dracut_modules():
    return list(api.consume(UpgradeDracutModule))


def install_initram_deps(context):
    used_repos = api.consume(UsedTargetRepositories)
    target_userspace_info = next(api.consume(TargetUserSpaceInfo), None)

    packages = set()
    for message in api.consume(RequiredUpgradeInitramPackages):
        packages.update(message.packages)
    dnfplugin.install_initramdisk_requirements(
        packages=packages, target_userspace_info=target_userspace_info, used_repos=used_repos)


def _install_files(context):
    """
    Copy the required files into the `context` userspace and return the list
    of those files.

    Those files will be copied from the source to the userspace used for
    generating of the upgrade initrd. These files may need to be installed
    in the initrd explicitly, so return the list of those files for other
    purposes.
    """
    # TODO: currently we need this just for /etc/dasd.conf, but it's expected
    # we will need this for more files in future. The concept used here will
    # need to be changed, as we should consume specific messages instead.
    # But for now this should be enough. Keeping that for future.
    if architecture.matches_architecture(architecture.ARCH_S390X):
        # we don't need to check existence of the file - it is required on this
        # this architecture
        context.copy_to('/etc/dasd.conf', '/etc/dasd.conf')
        return ['/etc/dasd.conf']
    return []


def install_multipath_files(context):
    # Include multipath related files (according to module-setup of multipath)
    if os.path.exists('/etc/xdrdevices.conf'):
        context.copy_to('/etc/xdrdevices.conf', '/etc/xdrdevices.conf')
    if os.path.exists('/etc/multipath.conf'):
        context.copy_to('/etc/multipath.conf', '/etc/multipath.conf')
        if os.path.isdir('/etc/multipath'):
            shutil.rmtree(context.full_path('/etc/multipath'))
            context.copytree_to('/etc/multipath', '/etc/multipath')


def generate_initram_disk(context):
    """
    Function to actually execute the initram creation.
    """
    generator_script = api.get_actor_file_path(INITRAM_GEN_SCRIPT_NAME)
    if not generator_script:
        raise StopActorExecutionError(
            message='Mandatory script to generate initram not available.',
            details=_reinstall_leapp_repository_hint()
        )
    modules = _get_dracut_modules()
    copy_dracut_modules(context, modules)
    context.copy_to(generator_script, os.path.join('/', INITRAM_GEN_SCRIPT_NAME))
    install_initram_deps(context)
    install_multipath_files(context)
    install_files = _install_files(context)
    # FIXME: issue #376
    context.call([
        '/bin/sh', '-c',
        'LEAPP_ADD_DRACUT_MODULES="{modules}" LEAPP_KERNEL_ARCH={arch} '
        'LEAPP_DRACUT_INSTALL_FILES="{files}" {cmd}'.format(
            modules=','.join([mod.name for mod in modules]),
            arch=api.current_actor().configuration.architecture,
            files=' '.join(install_files),
            cmd=os.path.join('/', INITRAM_GEN_SCRIPT_NAME))
    ])
    copy_boot_files(context)


def copy_boot_files(context):
    """
    Function to copy the generated initram and corresponding kernel to /boot - Additionally produces a BootContent
    message with their location.
    """
    curr_arch = api.current_actor().configuration.architecture
    kernel = 'vmlinuz-upgrade.{}'.format(curr_arch)
    initram = 'initramfs-upgrade.{}.img'.format(curr_arch)
    content = BootContent(
        kernel_path=os.path.join('/boot', kernel),
        initram_path=os.path.join('/boot', initram)
    )

    context.copy_from(os.path.join('/artifacts', kernel), content.kernel_path)
    context.copy_from(os.path.join('/artifacts', initram), content.initram_path)

    api.produce(content)


def process():
    userspace_info = next(api.consume(TargetUserSpaceInfo), None)

    with mounting.NspawnActions(base_dir=userspace_info.path) as context:
        generate_initram_disk(context)



from leapp.models import Model, fields
from leapp.topics import TransactionTopic


class RepositoryMap(Model):
    """
    Mapping between repositories to be used during upgrade.

    Mapping of current system repository to target system version repositories to determine which
    ones should be enabled for the correct upgrade process.
    """

    topic = TransactionTopic

    from_repoid = fields.String()
    """RHEL 7 repoid as present in the Red Hat CDN"""
    to_repoid = fields.String()
    """RHEL 8 repoid as present in the Red Hat CDN"""
    to_pes_repo = fields.String()
    """RHEL 8 repo name as used in the Package Evolution Service database"""
    from_minor_version = fields.String()
    """To which RHEL 7 minor versions the mapping relates to"""
    to_minor_version = fields.String()
    """To which RHEL 8 minor versions the mapping relates to"""
    arch = fields.String()
    """CPU architecture the mapping relates to"""
    repo_type = fields.StringEnum(choices=['rpm', 'srpm', 'debuginfo'])
    """Type of content the mapped repos hold"""


class RepositoriesMap(Model):
    topic = TransactionTopic

    repositories = fields.List(fields.Model(RepositoryMap), default=[])
    """List of RHEL 7 repo <-> RHEL 8 repo mappings"""



from leapp.exceptions import RequestStopAfterPhase
from leapp.libraries.stdlib import api
from leapp.reporting import Report


def check():
    if [msg for msg in api.consume(Report) if 'inhibitor' in msg.report.get('flags', [])]:
        raise RequestStopAfterPhase()



from leapp.libraries.stdlib import api, run, CalledProcessError
from leapp.models import SELinuxModules


def remove_semanage_customizations():
    # remove SELinux customizations done by semanage -- to be reintroduced after the upgrade
    api.current_logger().info('Removing SELinux customizations introduced by semanage.')

    semanage_options = ["login", "user", "port", "interface", "module", "node",
                        "fcontext", "boolean", "ibpkey", "ibendport"]
    # permissive domains are handled by porting modules (permissive -a adds new cil module with priority 400)
    for option in semanage_options:
        try:
            run(['semanage', option, '-D'])
        except CalledProcessError:
            continue


def remove_custom_modules():
    # remove custom SElinux modules - to be reinstalled after the upgrade
    for semodules in api.consume(SELinuxModules):
        api.current_logger().info("Removing custom SELinux policy modules. Count: %d", len(semodules.modules))
        for module in semodules.modules:
            api.current_logger().info("Removing %s on priority %d.", module.name, module.priority)
            try:
                run(['semodule',
                     '-X', str(module.priority),
                     '-r', module.name
                     ]
                    )
            except CalledProcessError as e:
                api.current_logger().warning("Failed to remove module %s on priority %d: %s",
                                             module.name, module.priority, str(e.stderr))
                continue



from leapp.actors import Actor
from leapp.models import StorageInfo
from leapp.reporting import Report, create_report
from leapp import reporting
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class InhibitWhenLuks(Actor):
    """
    Check if any encrypted partitions is in use. If yes, inhibit the upgrade process.

    Upgrading system with encrypted partition is not supported.
    """

    name = 'check_luks_and_inhibit'
    consumes = (StorageInfo,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        for storage_info in self.consume(StorageInfo):
            for blk in storage_info.lsblk:
                if blk.tp == 'crypt':
                    create_report([
                        reporting.Title('LUKS encrypted partition detected'),
                        reporting.Summary('Upgrading system with encrypted partitions is not supported'),
                        reporting.Severity(reporting.Severity.HIGH),
                        reporting.Tags([reporting.Tags.BOOT, reporting.Tags.ENCRYPTION]),
                        reporting.Flags([reporting.Flags.INHIBITOR]),
                    ])
                    break



from leapp.actors import Actor
from leapp.libraries.actor import scancpu
from leapp.models import CPUInfo
from leapp.tags import IPUWorkflowTag, FactsPhaseTag


class ScanCPU(Actor):
    """Scan CPUs of the machine."""

    name = 'scancpu'
    consumes = ()
    produces = (CPUInfo,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        scancpu.process()



from leapp.actors import Actor
from leapp.libraries.actor.grubdevname import get_grub_device
from leapp.libraries.common.config import architecture
from leapp.models import GrubDevice
from leapp.tags import FactsPhaseTag, IPUWorkflowTag
from leapp.utils.deprecation import suppress_deprecation


# TODO: remove this actor completely after the deprecation period expires
@suppress_deprecation(GrubDevice)
class Grubdevname(Actor):
    """
    Get name of block device where GRUB is located
    """

    name = 'grubdevname'
    consumes = ()
    produces = (GrubDevice,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        if architecture.matches_architecture(architecture.ARCH_S390X):
            return
        get_grub_device()



from leapp.libraries.stdlib import api, run
from leapp.models import StorageInfo, XFSPresence


def scan_xfs_fstab(data):
    mountpoints = set()
    for entry in data:
        if entry.fs_vfstype == "xfs":
            mountpoints.add(entry.fs_file)

    return mountpoints


def scan_xfs_mount(data):
    mountpoints = set()
    for entry in data:
        if entry.tp == "xfs":
            mountpoints.add(entry.mount)

    return mountpoints


def scan_xfs_systemdmount(data):
    mountpoints = set()
    for entry in data:
        if entry.fs_type == "xfs":
            mountpoints.add(entry.path)

    return mountpoints


def is_xfs_without_ftype(mp):
    for l in run(['/usr/sbin/xfs_info', '{}'.format(mp)], split=True)['stdout']:
        if 'ftype=0' in l:
            return True

    return False


def scan_xfs():
    storage_info_msgs = api.consume(StorageInfo)
    storage_info = next(storage_info_msgs, None)

    if list(storage_info_msgs):
        api.current_logger().warning('Unexpectedly received more than one StorageInfo message.')

    fstab_data = set()
    mount_data = set()
    systemdmount_data = set()
    if storage_info:
        fstab_data = scan_xfs_fstab(storage_info.fstab)
        mount_data = scan_xfs_mount(storage_info.mount)
        systemdmount_data = scan_xfs_systemdmount(storage_info.systemdmount)

    mountpoints = fstab_data | mount_data | systemdmount_data
    mountpoints_ftype0 = list(filter(is_xfs_without_ftype, mountpoints))

    # By now, we only have XFS mountpoints and check whether or not it has ftype = 0
    api.produce(XFSPresence(
        present=len(mountpoints) > 0,
        without_ftype=len(mountpoints_ftype0) > 0,
        mountpoints_without_ftype=mountpoints_ftype0,
    ))



import os

from leapp.libraries.stdlib import api


AWS_MAP = {
    'el7_pkg': 'rh-amazon-rhui-client',
    'el8_pkg': 'rh-amazon-rhui-client',
    'leapp_pkg': 'leapp-rhui-aws',
    'leapp_pkg_repo': 'leapp-aws.repo'
}

AZURE_MAP = {
    'el7_pkg': 'rhui-azure-rhel7',
    'el8_pkg': 'rhui-azure-rhel8',
    'agent_pkg': 'WALinuxAgent',
    'leapp_pkg': 'leapp-rhui-azure',
    'leapp_pkg_repo': 'leapp-azure.repo'
}

# for the moment the only difference in RHUI package naming is on ARM
AWS_MAP_AARCH64 = dict(AWS_MAP, el7_pkg='rh-amazon-rhui-client-arm')

RHUI_CLOUD_MAP = {
    'x86_64': {
        'aws': AWS_MAP,
        'azure': AZURE_MAP,
    },
    'aarch64': {
        'aws': AWS_MAP_AARCH64,
        'azure': AZURE_MAP,
    },
    'ppc64le': {
        'aws': AWS_MAP,
        'azure': AZURE_MAP,
    },
    's390x': {
        'aws': AWS_MAP,
        'azure': AZURE_MAP,
    },
}

# when on AWS, we need also Python2 version of "Amazon-id" dnf plugin which is served by
# "leapp-rhui-aws" rpm package (please note this package is not in any RH official repository
# but only in "rhui-client-config-*" repo)
DNF_PLUGIN_PATH = '/usr/lib/python2.7/site-packages/dnf-plugins/'
YUM_REPOS_PATH = '/etc/yum.repos.d'

RHUI_PKI_DIR = '/etc/pki/rhui'
RHUI_PKI_PRODUCT_DIR = os.path.join(RHUI_PKI_DIR, 'product')
RHUI_PKI_PRIVATE_DIR = os.path.join(RHUI_PKI_DIR, 'private')

AWS_DNF_PLUGIN_NAME = 'amazon-id.py'

# these files are provided by special Leapp rpms (per cloud) and
# are delivered into "repos/system_upgrade/el7toel8/files/rhui/<PROVIDER>


def gen_rhui_files_map():

    arch = api.current_actor().configuration.architecture

    return {
        'aws': [
            ('rhui-client-config-server-8.crt', RHUI_PKI_PRODUCT_DIR),
            ('rhui-client-config-server-8.key', RHUI_PKI_DIR),
            ('content-rhel8.crt', RHUI_PKI_PRODUCT_DIR),
            ('content-rhel8.key', RHUI_PKI_DIR),
            ('cdn.redhat.com-chain.crt', RHUI_PKI_DIR),
            (AWS_DNF_PLUGIN_NAME, DNF_PLUGIN_PATH),
            (RHUI_CLOUD_MAP[arch]['aws']['leapp_pkg_repo'], YUM_REPOS_PATH)
        ],
        'azure': [
            ('content.crt', RHUI_PKI_PRODUCT_DIR),
            ('key.pem', RHUI_PKI_PRIVATE_DIR),
            (RHUI_CLOUD_MAP[arch]['azure']['leapp_pkg_repo'], YUM_REPOS_PATH)
        ]
    }


def copy_rhui_data(context, provider):

    rhui_dir = api.get_common_folder_path('rhui')
    data_dir = os.path.join(rhui_dir, provider)

    context.call(['mkdir', '-p', RHUI_PKI_PRODUCT_DIR])
    context.call(['mkdir', '-p', RHUI_PKI_PRIVATE_DIR])

    for path_ in gen_rhui_files_map().get(provider, ()):
        context.copy_to(os.path.join(data_dir, path_[0]), path_[1])



import os

from leapp import reporting
from leapp.actors import Actor
from leapp.exceptions import StopActorExecutionError
from leapp.models import Report, RootDirectory
from leapp.tags import IPUWorkflowTag, ChecksPhaseTag


class CheckRootSymlinks(Actor):
    """
    Check if the symlinks /bin and /lib are relative, not absolute.

    After reboot, dracut fails if the links are absolute.
    """

    name = 'check_root_symlinks'
    consumes = (RootDirectory,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        rootdir = next(self.consume(RootDirectory), None)
        if not rootdir:
            raise StopActorExecutionError('Cannot check root symlinks',
                                          details={'Problem': 'Did not receive a message with '
                                                              'root subdirectories'})
        absolute_links = [item for item in rootdir.items if item.target and os.path.isabs(item.target)]

        if absolute_links:
            commands = [' '.join(['ln', '-snf',
                                  os.path.relpath(item.target, '/'),
                                  os.path.join('/', item.name)]) for item in absolute_links]
            remediation = [['sh', '-c', ' && '.join(commands)]]
            reporting.create_report([
                reporting.Title('Upgrade requires links in root directory to be relative'),
                reporting.Summary(
                    'After rebooting, parts of the upgrade process can fail if symbolic links in / '
                    'point to absolute paths.\n'
                    'Please change these links to relative ones.'
                ),
                reporting.Severity(reporting.Severity.HIGH),
                reporting.Flags([reporting.Flags.INHIBITOR]),
                reporting.Remediation(commands=remediation)
            ])



from collections import namedtuple

import pytest

from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import repofileutils, rhsm
from leapp.libraries.common.testutils import create_report_mocked, logger_mocked
from leapp.libraries.stdlib import CalledProcessError, api
from leapp.models import RepositoryFile, RepositoryData

Repository = namedtuple('Repository', ['repoid', 'file'])
LIST_SEPARATOR = '\n    - '


class IsolatedActionsMocked(object):
    def __init__(self, call_stdout=None, raise_err=False):
        self.commands_called = []
        self.call_return = {'stdout': call_stdout, 'stderr': None}
        self.raise_err = raise_err

    def call(self, cmd, *args):
        self.commands_called.append(cmd)
        if self.raise_err:
            raise_call_error(cmd)
        return self.call_return


def raise_call_error(args=None, exit_code=1):
    raise CalledProcessError(
        message='Command {0} failed with exit code {1}.'.format(str(args), exit_code),
        command=args,
        result={'signal': None, 'exit_code': exit_code, 'pid': 0, 'stdout': 'fake out', 'stderr': 'fake err'}
    )


def _gen_repo(repoid):
    return RepositoryData(repoid=repoid, name='name {}'.format(repoid))


def _gen_repofile(rfile, data=None):
    if data is None:
        data = [_gen_repo("{}-{}".format(rfile.split("/")[-1], i)) for i in range(3)]
    return RepositoryFile(file=rfile, data=data)


@pytest.mark.parametrize('other_repofiles', [
    [],
    [_gen_repofile("foo")],
    [_gen_repofile("foo"), _gen_repofile("bar")],
])
@pytest.mark.parametrize('rhsm_repofile', [
    None,
    _gen_repofile(rhsm._DEFAULT_RHSM_REPOFILE, []),
    _gen_repofile(rhsm._DEFAULT_RHSM_REPOFILE, [_gen_repo("rh-0")]),
    _gen_repofile(rhsm._DEFAULT_RHSM_REPOFILE),
])
def test_get_available_repo_ids(monkeypatch, other_repofiles, rhsm_repofile):
    context_mocked = IsolatedActionsMocked()
    repos = other_repofiles[:]
    if rhsm_repofile:
        repos.append(rhsm_repofile)
    rhsm_repos = [repo.repoid for repo in rhsm_repofile.data] if rhsm_repofile else []

    monkeypatch.setattr(api, 'current_logger', logger_mocked())
    monkeypatch.setattr(rhsm, '_inhibit_on_duplicate_repos', lambda x: None)
    monkeypatch.setattr(repofileutils, 'get_parsed_repofiles', lambda x: repos)

    result = rhsm.get_available_repo_ids(context_mocked)

    rhsm_repos.sort()
    assert context_mocked.commands_called == [['yum', 'clean', 'all']]
    assert result == rhsm_repos
    if result:
        msg = (
            'The following repoids are available through RHSM:{0}{1}'
            .format(LIST_SEPARATOR, LIST_SEPARATOR.join(rhsm_repos))
        )
        assert msg in api.current_logger.infomsg
    else:
        assert 'There are no repos available through RHSM.' in api.current_logger.infomsg


def test_get_available_repo_ids_error():
    context_mocked = IsolatedActionsMocked(raise_err=True)

    with pytest.raises(StopActorExecutionError) as err:
        rhsm.get_available_repo_ids(context_mocked)

    assert 'Unable to use yum' in str(err)


def test_inhibit_on_duplicate_repos(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(api, 'current_logger', logger_mocked())
    repofiles = [
        _gen_repofile("foo", [_gen_repo('repoX'), _gen_repo('repoY')]),
        _gen_repofile("bar", [_gen_repo('repoX')]),
    ]

    rhsm._inhibit_on_duplicate_repos(repofiles)

    dups = ['repoX']
    assert ('The following repoids are defined multiple times:{0}{1}'
            .format(LIST_SEPARATOR, LIST_SEPARATOR.join(dups))) in api.current_logger.warnmsg
    assert reporting.create_report.called == 1
    assert 'inhibitor' in reporting.create_report.report_fields['flags']
    assert reporting.create_report.report_fields['title'] == 'A YUM/DNF repository defined multiple times'
    summary = ('The following repositories are defined multiple times:{0}{1}'
               .format(LIST_SEPARATOR, LIST_SEPARATOR.join(dups)))
    assert summary in reporting.create_report.report_fields['summary']


def test_inhibit_on_duplicate_repos_no_dups(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(api, 'current_logger', logger_mocked())

    rhsm._inhibit_on_duplicate_repos([_gen_repofile("foo")])

    assert not api.current_logger.warnmsg
    assert reporting.create_report.called == 0



import os

import pytest

from leapp.exceptions import StopActorExecution
from leapp.libraries.actor import grubdevname
from leapp.libraries.common import testutils
from leapp.libraries.stdlib import CalledProcessError, api

BOOT_PARTITION = '/dev/vda1'

BOOT_DEVICE = '/dev/vda'
BOOT_DEVICE_ENV = '/dev/sda'

VALID_DD = b'GRUB GeomHard DiskRead Error'
INVALID_DD = b'Nothing here'

CUR_DIR = os.path.dirname(os.path.abspath(__file__))


def raise_call_error(args=None):
    raise CalledProcessError(
        message='A Leapp Command Error occured.',
        command=args,
        result={'signal': None, 'exit_code': 1, 'pid': 0, 'stdout': 'fake', 'stderr': 'fake'}
    )


class RunMocked(object):

    def __init__(self, raise_err=False):
        self.called = 0
        self.args = None
        self.raise_err = raise_err

    def __call__(self, args, encoding=None):
        self.called += 1
        self.args = args
        if self.raise_err:
            raise_call_error(args)

        if self.args == ['grub2-probe', '--target=device', '/boot']:
            stdout = BOOT_PARTITION

        elif self.args == ['lsblk', '-spnlo', 'name', BOOT_PARTITION]:
            stdout = BOOT_DEVICE

        return {'stdout': stdout}


def open_mocked(fn, flags):
    return open(os.path.join(CUR_DIR, 'valid') if fn == BOOT_DEVICE else os.path.join(CUR_DIR, 'invalid'), 'r')


def open_invalid(fn, flags):
    return open(os.path.join(CUR_DIR, 'invalid'), 'r')


def read_mocked(f, size):
    return f.read(size)


def close_mocked(f):
    f.close()


def test_get_grub_device(monkeypatch):
    run_mocked = RunMocked()
    monkeypatch.setattr(grubdevname, 'run', run_mocked)
    monkeypatch.setattr(api, 'produce', testutils.produce_mocked())
    monkeypatch.setattr(os, 'open', open_mocked)
    monkeypatch.setattr(os, 'read', read_mocked)
    monkeypatch.setattr(os, 'close', close_mocked)
    grubdevname.get_grub_device()
    assert grubdevname.run.called == 2
    assert BOOT_DEVICE == api.produce.model_instances[0].grub_device


def test_get_grub_device_fail(monkeypatch):
    run_mocked = RunMocked(raise_err=True)
    monkeypatch.setattr(grubdevname, 'run', run_mocked)
    monkeypatch.setattr(api, 'produce', testutils.produce_mocked())
    monkeypatch.setattr(os, 'open', open_mocked)
    monkeypatch.setattr(os, 'read', read_mocked)
    monkeypatch.setattr(os, 'close', close_mocked)
    with pytest.raises(StopActorExecution):
        grubdevname.get_grub_device()
    assert grubdevname.run.called == 1
    assert not api.produce.model_instances


def test_grub_device_env_var(monkeypatch):
    run_mocked = RunMocked()
    monkeypatch.setenv('LEAPP_GRUB_DEVICE', BOOT_DEVICE_ENV)
    monkeypatch.setattr(grubdevname, 'run', run_mocked)
    monkeypatch.setattr(api, 'produce', testutils.produce_mocked())
    monkeypatch.setattr(os, 'open', open_mocked)
    monkeypatch.setattr(os, 'read', read_mocked)
    monkeypatch.setattr(os, 'close', close_mocked)
    grubdevname.get_grub_device()
    assert grubdevname.run.called == 0
    assert BOOT_DEVICE_ENV == api.produce.model_instances[0].grub_device


def test_device_no_grub(monkeypatch):
    run_mocked = RunMocked()
    monkeypatch.setattr(grubdevname, 'run', run_mocked)
    monkeypatch.setattr(api, 'produce', testutils.produce_mocked())
    monkeypatch.setattr(os, 'open', open_invalid)
    monkeypatch.setattr(os, 'read', read_mocked)
    monkeypatch.setattr(os, 'close', close_mocked)
    grubdevname.get_grub_device()
    assert grubdevname.run.called == 2
    assert not api.produce.model_instances



from leapp.actors import Actor
from leapp.libraries.actor.repositoriesblacklist import process
from leapp.models import RepositoriesBlacklisted, RepositoriesFacts, RepositoriesMap
from leapp.reporting import Report
from leapp.tags import IPUWorkflowTag, FactsPhaseTag


class RepositoriesBlacklist(Actor):
    """
    Generate list of repository IDs that should be ignored by Leapp during upgrade process
    """

    name = 'repositories_blacklist'
    consumes = (RepositoriesFacts, RepositoriesMap, )
    produces = (RepositoriesBlacklisted, Report)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        process()



import json
import os
from collections import namedtuple, defaultdict
from enum import IntEnum

from leapp.exceptions import StopActorExecution, StopActorExecutionError
from leapp import reporting
from leapp.libraries.common.config import architecture, version
from leapp.libraries.common import fetch
from leapp.libraries.stdlib import api
from leapp.libraries.stdlib.config import is_verbose
from leapp.models import (InstalledRedHatSignedRPM, PESRpmTransactionTasks, RepositoriesMap,
                          RepositoriesSetupTasks, RpmTransactionTasks, RepositoriesBlacklisted)


Event = namedtuple('Event', ['id',            # int
                             'action',        # An instance of Action
                             'in_pkgs',       # A dictionary with packages in format {<pkg_name>: <repository>}
                             'out_pkgs',      # A dictionary with packages in format {<pkg_name>: <repository>}
                             'from_release',  # A tuple representing a release in format (major, minor)
                             'to_release',    # A tuple representing a release in format (major, minor)
                             'architectures'  # A list of strings representing architectures
                             ])


class Action(IntEnum):
    PRESENT = 0
    REMOVED = 1
    DEPRECATED = 2
    REPLACED = 3
    SPLIT = 4
    MERGED = 5
    MOVED = 6
    RENAMED = 7


class Task(IntEnum):
    KEEP = 0
    INSTALL = 1
    REMOVE = 2

    def past(self):
        return ['kept', 'installed', 'removed'][self]


def pes_events_scanner(pes_json_directory, pes_json_filename):
    """Entrypoint to the library"""
    installed_pkgs = get_installed_pkgs()
    transaction_configuration = get_transaction_configuration()
    arch = api.current_actor().configuration.architecture
    events = get_events(pes_json_directory, pes_json_filename)
    releases = get_releases(events)
    target = version._version_to_tuple(api.current_actor().configuration.version.target)

    filtered_releases = filter_releases_by_target(releases, target)
    filtered_events = filter_events_by_releases(events, filtered_releases)
    arch_events = filter_events_by_architecture(filtered_events, arch)

    add_output_pkgs_to_transaction_conf(transaction_configuration, arch_events)
    drop_conflicting_release_events(arch_events)
    tasks = process_events(filtered_releases, arch_events, installed_pkgs)
    filter_out_transaction_conf_pkgs(tasks, transaction_configuration)
    produce_messages(tasks)


def get_installed_pkgs():
    """
    Get installed Red Hat-signed packages.

    :return: Set of names of the installed Red Hat-signed packages
    """
    installed_pkgs = set()

    installed_rh_signed_rpm_msgs = api.consume(InstalledRedHatSignedRPM)
    installed_rh_signed_rpm_msg = next(installed_rh_signed_rpm_msgs, None)
    if list(installed_rh_signed_rpm_msgs):
        api.current_logger().warning('Unexpectedly received more than one InstalledRedHatSignedRPM message.')
    if not installed_rh_signed_rpm_msg:
        raise StopActorExecutionError('Cannot parse PES data properly due to missing list of installed packages',
                                      details={'Problem': 'Did not receive a message with installed Red Hat-signed '
                                                          'packages (InstalledRedHatSignedRPM)'})
    installed_pkgs.update([pkg.name for pkg in installed_rh_signed_rpm_msg.items])
    return installed_pkgs


def _get_repositories_mapping():
    """
    Get all repositories mapped from repomap file and map repositories id with respective names.

    :return: Dictionary with all repositories mapped.
    """
    repositories_mapping = {}

    repositories_map_msgs = api.consume(RepositoriesMap)
    repositories_map_msg = next(repositories_map_msgs, None)
    if list(repositories_map_msgs):
        api.current_logger().warning('Unexpectedly received more than one RepositoriesMap message.')
    if not repositories_map_msg:
        raise StopActorExecutionError(
            'Cannot parse RepositoriesMap data properly',
            details={'Problem': 'Did not receive a message with mapped repositories'}
        )

    for repository in repositories_map_msg.repositories:
        if repository.arch == api.current_actor().configuration.architecture:
            repositories_mapping[repository.to_pes_repo] = repository.to_repoid

    return repositories_mapping


def get_transaction_configuration():
    """
    Get pkgs to install, keep and remove from the user configuration files in /etc/leapp/transaction/.

    These configuration files have higher priority than PES data.
    :return: RpmTransactionTasks model instance
    """
    transaction_configuration = RpmTransactionTasks()

    for tasks in api.consume(RpmTransactionTasks):
        transaction_configuration.to_install.extend(tasks.to_install)
        transaction_configuration.to_remove.extend(tasks.to_remove)
        transaction_configuration.to_keep.extend(tasks.to_keep)
    return transaction_configuration


def filter_releases_by_target(releases, target):
    match_list = ['<= {}.{}'.format(*target)]
    return [r for r in releases if version.matches_version(match_list, '{}.{}'.format(*r))]


def get_events(pes_json_directory, pes_json_filename):
    """
    Get all the events from the source JSON file exported from PES.

    :return: List of Event tuples, where each event contains event type and input/output pkgs
    """
    try:
        return parse_pes_events(
            fetch.read_or_fetch(pes_json_filename, directory=pes_json_directory, allow_empty=True))
    except (ValueError, KeyError):
        title = 'Missing/Invalid PES data file ({}/{})'.format(pes_json_directory, pes_json_filename)
        summary = 'Read documentation at: https://access.redhat.com/articles/3664871 for more information ' \
            'about how to retrieve the files'
        reporting.create_report([
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Tags([reporting.Tags.SANITY]),
            reporting.Flags([reporting.Flags.INHIBITOR]),
            reporting.RelatedResource('file', os.path.join(pes_json_directory, pes_json_filename))
        ])
        raise StopActorExecution()


def get_releases(events):
    """
    Get all target releases from a list of events.

    :return: List of (major, minor) release tuples, sorted in ascending order
    """
    return sorted({event.to_release for event in events})


def filter_events_by_architecture(events, arch):
    return [e for e in events if not e.architectures or arch in e.architectures]


def filter_events_by_releases(events, releases):
    return [e for e in events if e.to_release in releases]


def parse_pes_events(json_data):
    """
    Parse JSON data returning PES events

    :return: List of Event tuples, where each event contains event type and input/output pkgs
    """
    data = json.loads(json_data)
    if not isinstance(data, dict) or not data.get('packageinfo'):
        raise ValueError('Found PES data with invalid structure')

    return [parse_entry(entry) for entry in data['packageinfo']]


def parse_entry(entry):
    """
    Parse PES event data

    :param entry: A dict with the following structure:
                  {
                      'action': 3,
                      'id': 15,
                      'initial_release': {  # can be None
                          'z_stream': None,
                          'major_version': 7,
                          'tag': None,
                          'os_name': 'RHEL',
                          'minor_version': 7
                      },
                      'release': {
                          'z_stream': None,
                          'major_version': 8,
                          'tag': None,
                          'os_name': 'RHEL',
                          'minor_version': 0
                      },
                      'in_packageset': {
                          'set_id': 20,
                          'package': [
                              {
                                  'name': 'espeak',
                                  'repository': 'rhel7-optional'
                              }
                          ]
                      },
                      'out_packageset': {  # can be None
                          'set_id': 21,
                          'package': [
                              {
                                  'name': 'espeak-ng',
                                  'repository': 'rhel8-AppStream'
                              }
                          ]
                      },
                      'architectures': [  # can be empty
                          'x86_64',
                          'aarch64',
                          'ppc64le',
                          's390x'
                      ]
                  }
    """

    event_id = entry.get('id') or 0
    action = parse_action(entry['action'])
    in_pkgs = parse_packageset(entry.get('in_packageset') or {})
    out_pkgs = parse_packageset(entry.get('out_packageset') or {})
    from_release = parse_release(entry.get('initial_release') or {})
    to_release = parse_release(entry.get('release') or {})
    architectures = parse_architectures(entry.get('architectures') or [])

    return Event(event_id, action, in_pkgs, out_pkgs, from_release, to_release, architectures)


def parse_action(action_id):
    """Get event type name based on PES event's action id"""
    if action_id < 0 or action_id >= len(Action):
        raise ValueError('Found event with invalid action ID: {}'. format(action_id))
    return Action(action_id)


def parse_packageset(packageset):
    """
    Get "input" or "output" packages and their repositories from each PES event.

    :return: A dictionary with packages in format {<pkg_name>: <repository>}
    """
    return {p['name']: p['repository'].lower() for p in packageset.get('package', [])}


def parse_release(release):
    return (release['major_version'], release['minor_version']) if release else (0, 0)


def parse_architectures(architectures):
    for arch in architectures:
        if arch not in architecture.ARCH_ACCEPTED:
            raise ValueError('Found event with invalid architecture: {}'. format(arch))
    return architectures


def is_event_relevant(event, installed_pkgs, tasks):
    """Determine if event is applicable given the installed packages and tasks planned so far."""
    for package in event.in_pkgs.keys():
        if package in tasks[Task.REMOVE] and event.action != Action.PRESENT:
            return False
        if package not in installed_pkgs and package not in tasks[Task.INSTALL]:
            return False
    return True


def add_packages_to_tasks(tasks, packages, task_type):
    if packages:
        api.current_logger().debug('{v:7} {p}'.format(v=task_type.name.upper(), p=', '.join(packages)))
        tasks[task_type].update(packages)


def _packages_to_str(packages):
    """
    Represent a dictionary holding a set of packages with a string hash.

    Example: in: {'mesa-libwayland-egl-devel': 'rhel7-optional', 'wayland-devel': 'rhel7-base'}
             out: 'mesa-libwayland-egl-devel:rhel7-optional,wayland-devel:rhel7-base'
    """
    return ','.join('{}:{}'.format(p[0], p[1]) for p in sorted(packages.items()))


def drop_conflicting_release_events(events):
    """
    In case of events with identical initial release and input packages, drop those with older target releases.

    In case of identical target releases too, drop events with lower IDs.
    """
    events_by_input = defaultdict(list)  # {(release, input packages): [event]}
    for event in events:
        input_packages_str = _packages_to_str(event.in_pkgs)
        events_by_input[(event.from_release, input_packages_str)].append(event)
    for input_events in events_by_input.values():
        if len(input_events) > 1:
            input_events.sort(key=lambda e: (e.to_release, e.id))
            api.current_logger().debug('Conflicting events with same input packages and initial release: #' +
                                       ', #'.join(str(e.id) for e in input_events))
            for event in input_events[:-1]:
                api.current_logger().debug('Dropping event #{}'.format(event.id))
                events.remove(event)


def process_events(releases, events, installed_pkgs):
    """
    Process PES events to get lists of pkgs to keep, to install and to remove.

    :param releases: List of tuples representing ordered releases in format (major, minor)
    :param events: List of Event tuples, not including those events with their "input" packages not installed
    :param installed_pkgs: Set of names of the installed Red Hat-signed packages
    :return: A dict with three dicts holding pkgs to keep, to install and to remove
    """
    # subdicts in format {<pkg_name>: <repository>}
    tasks = {t: {} for t in Task}  # noqa: E1133; pylint: disable=not-an-iterable

    for release in releases:
        current = {t: {} for t in Task}  # noqa: E1133; pylint: disable=not-an-iterable
        release_events = [e for e in events if e.to_release == release]
        api.current_logger().debug('---- Processing {n} eligible events for release {r}'.format(
            n=len(release_events), r=release))

        for event in release_events:
            if is_event_relevant(event, installed_pkgs, tasks):
                if event.action in [Action.DEPRECATED, Action.PRESENT]:
                    # Keep these packages to make sure the repo they're in on RHEL 8 gets enabled
                    add_packages_to_tasks(current, event.in_pkgs, Task.KEEP)

                if event.action == Action.MOVED:
                    # Keep these packages to make sure the repo they're in on RHEL 8 gets enabled
                    # We don't care about the "in_pkgs" as it contains always just one pkg - the same as the "out" pkg
                    add_packages_to_tasks(current, event.out_pkgs, Task.KEEP)

                if event.action in [Action.SPLIT, Action.MERGED, Action.RENAMED, Action.REPLACED]:
                    non_installed_out_pkgs = filter_out_installed_pkgs(event.out_pkgs, installed_pkgs)
                    add_packages_to_tasks(current, non_installed_out_pkgs, Task.INSTALL)
                    # Keep already installed "out" pkgs to ensure the repo they're in on RHEL 8 gets enabled
                    installed_out_pkgs = get_installed_event_pkgs(event.out_pkgs, installed_pkgs)
                    add_packages_to_tasks(current, installed_out_pkgs, Task.KEEP)

                    if event.action in [Action.SPLIT, Action.MERGED]:
                        # Remove those RHEL 7 pkgs that are no longer on RHEL 8
                        in_pkgs_without_out_pkgs = filter_out_out_pkgs(event.in_pkgs, event.out_pkgs)
                        add_packages_to_tasks(current, in_pkgs_without_out_pkgs, Task.REMOVE)

                if event.action in [Action.RENAMED, Action.REPLACED, Action.REMOVED]:
                    add_packages_to_tasks(current, event.in_pkgs, Task.REMOVE)

        do_not_remove = set()
        for package in current[Task.REMOVE]:
            if package in tasks[Task.KEEP]:
                api.current_logger().warning(
                    '{p} :: {r} to be kept / currently removed - removing package'.format(
                        p=package, r=current[Task.REMOVE][package]))
                del tasks[Task.KEEP][package]
            elif package in tasks[Task.INSTALL]:
                api.current_logger().warning(
                    '{p} :: {r} to be installed / currently removed - ignoring tasks'.format(
                        p=package, r=current[Task.REMOVE][package]))
                del tasks[Task.INSTALL][package]
                do_not_remove.add(package)
        for package in do_not_remove:
            del current[Task.REMOVE][package]

        do_not_install = set()
        for package in current[Task.INSTALL]:
            if package in tasks[Task.REMOVE]:
                api.current_logger().warning(
                    '{p} :: {r} to be removed / currently installed - keeping package'.format(
                        p=package, r=current[Task.INSTALL][package]))
                current[Task.KEEP][package] = current[Task.INSTALL][package]
                del tasks[Task.REMOVE][package]
                do_not_install.add(package)
        for package in do_not_install:
            del current[Task.INSTALL][package]

        for package in current[Task.KEEP]:
            if package in tasks[Task.REMOVE]:
                api.current_logger().warning(
                    '{p} :: {r} to be removed / currently kept - keeping package'.format(
                        p=package, r=current[Task.KEEP][package]))
                del tasks[Task.REMOVE][package]

        for key in Task:  # noqa: E1133; pylint: disable=not-an-iterable
            for package in current[key]:
                if package in tasks[key]:
                    api.current_logger().warning(
                        '{p} :: {r} to be {v} TWICE - internal bug (not serious, continuing)'.format(
                            p=package, r=current[key][package], v=key.past()))
            tasks[key].update(current[key])

    map_repositories(tasks[Task.INSTALL])
    map_repositories(tasks[Task.KEEP])
    filter_out_pkgs_in_blacklisted_repos(tasks[Task.INSTALL])
    resolve_conflicting_requests(tasks)

    return tasks


def filter_out_installed_pkgs(event_out_pkgs, installed_pkgs):
    """Do not install those packages that are already installed."""
    return {k: v for k, v in event_out_pkgs.items() if k not in installed_pkgs}


def get_installed_event_pkgs(event_pkgs, installed_pkgs):
    """
    Get those event's "in" or "out" packages which are already installed.

    Even though we don't want to install the already installed pkgs, to be able to upgrade them to their RHEL 8 version
    we need to know in which repos they are and enable such repos.
    """
    return {k: v for k, v in event_pkgs.items() if k in installed_pkgs}


def filter_out_out_pkgs(event_in_pkgs, event_out_pkgs):
    """
    In case of a Split or Merge PES events some of the "out" packages can be the same as the "in" packages.

    We don't want to remove those "in" packages that are among "out" packages. For example in case of a split of gdbm
    to gdbm and gdbm-libs, we would incorrectly mandate removing gdbm without this filter. But for example in case of
    a split of Cython to python2-Cython and python3-Cython, we will correctly mandate removing Cython.
    """
    return {k: v for k, v in event_in_pkgs.items() if k not in event_out_pkgs}


SKIPPED_PKGS_MSG = (
    'packages will be skipped because they are available only in '
    'RHEL 8 repositories that are intentionally excluded from the '
    'list of repositories used during the upgrade. '
    'See the report message titled "Excluded RHEL 8 repositories" '
    'for details.\nThe list of these packages:'
)


def filter_out_pkgs_in_blacklisted_repos(to_install):
    """
    Do not install packages that are available in blacklisted repositories

    No need to filter out the to_keep packages as the blacklisted repos won't get enabled - that is ensured in the
    setuptargetrepos actor. So even if they fall into the 'yum upgrade' bucket, they won't be available thus upgraded.
    """
    # FIXME The to_install contains just a limited subset of packages - those that are *not* currently installed and
    #   are to be installed. But we should also warn about the packages that *are* installed.
    blacklisted_pkg_repo_pairs = set()
    blacklisted_repos = get_repositories_blacklisted()
    for pkg, repo in to_install.items():
        if repo in blacklisted_repos:
            blacklisted_pkg_repo_pairs.add((pkg, repo))

    for pkg, _ in blacklisted_pkg_repo_pairs:
        del to_install[pkg]

    if blacklisted_pkg_repo_pairs:
        report_skipped_packages(
            title='Packages available in excluded repositories will not be installed',
            message=SKIPPED_PKGS_MSG,
            package_repo_pairs=blacklisted_pkg_repo_pairs,
        )


def resolve_conflicting_requests(tasks):
    """
    Do not remove what is supposed to be kept or installed.

    PES events may give us conflicting requests - to both install/keep and remove a pkg.
    Example of two real-world PES events resulting in a conflict:
      PES event 1: sip-devel  SPLIT INTO   python3-sip-devel, sip
      PES event 2: sip        SPLIT INTO   python3-pyqt5-sip, python3-sip
        -> without this function, sip would reside in both [Task.KEEP] and [Task.REMOVE], causing a dnf conflict
    """
    pkgs_in_conflict = set()
    for pkg in list(tasks[Task.INSTALL].keys()) + list(tasks[Task.KEEP].keys()):
        if pkg in tasks[Task.REMOVE]:
            pkgs_in_conflict.add(pkg)
            del tasks[Task.REMOVE][pkg]

    if pkgs_in_conflict:
        api.current_logger().debug('The following packages were marked to be kept/installed and removed at the same'
                                   ' time. Leapp will upgrade them.\n{}'.format('\n'.join(sorted(pkgs_in_conflict))))


def get_repositories_blacklisted():
    """Consume message and return a set of blacklisted repositories"""
    repos_blacklisted = set()
    for blacklist in api.consume(RepositoriesBlacklisted):
        repos_blacklisted.update(blacklist.repoids)
    return repos_blacklisted


def map_repositories(packages):
    """Map repositories from PES data to RHSM repository id"""
    repositories_mapping = _get_repositories_mapping()
    pkg_with_repo_without_mapping = set()
    for pkg, repo in packages.items():
        if repo not in repositories_mapping:
            pkg_with_repo_without_mapping.add((pkg, repo))
            continue

        packages[pkg] = repositories_mapping[repo]

    for pkg, _ in pkg_with_repo_without_mapping:
        del packages[pkg]

    if pkg_with_repo_without_mapping:
        report_skipped_packages(
            title='Packages from unknown repositories may not be installed',
            message='packages may not be installed or upgraded due to repositories unknown to leapp:',
            package_repo_pairs=pkg_with_repo_without_mapping,
            remediation=(
                "Please file a bug in http://bugzilla.redhat.com/ for leapp-repository component of "
                "the Red Hat Enterprise Linux 7 product."
            ),
        )


def report_skipped_packages(title, message, package_repo_pairs, remediation=None):
    """Generate report message about skipped packages"""
    package_repo_pairs = sorted(package_repo_pairs)
    summary = '{} {}\n{}'.format(
        len(package_repo_pairs), message, '\n'.join(
            [
                '- {pkg} (repoid: {repo})'.format(pkg=pkg, repo=repo)
                for pkg, repo in package_repo_pairs
            ]
        )
    )
    report_content = [
        reporting.Title(title),
        reporting.Summary(summary),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Tags([reporting.Tags.REPOSITORY]),
    ]
    if remediation:
        report_content += [reporting.Remediation(hint=remediation)]
    report_content += [reporting.RelatedResource('package', p) for p, _ in package_repo_pairs]
    reporting.create_report(report_content)
    if is_verbose():
        api.current_logger().info(summary)


def add_output_pkgs_to_transaction_conf(transaction_configuration, events):
    """
    Filter out those PES events that conflict with the higher priority transaction configuration files.

    Output packages from an event are added to packages for removal only if all input packages are already there.

    :param transaction_configuration: RpmTransactionTasks model instance with pkgs to install, keep and remove based
                                      on the user configuration files
    :param events: List of Event tuples, where each event contains event type and input/output pkgs
    """
    message = 'The following RHEL 8 packages will not be installed:\n'

    for event in events:
        if event.action in (Action.SPLIT, Action.MERGED, Action.REPLACED, Action.RENAMED):
            if all([pkg in transaction_configuration.to_remove for pkg in event.in_pkgs]):
                transaction_configuration.to_remove.extend(event.out_pkgs)
                message += (
                    '- {outs}\n - Reason: {ins} being {action} {to_or_by} {outs} {is_or_are} mentioned in the'
                    ' transaction configuration file /etc/leapp/transaction/to_remove\n'.format(
                        outs=', '.join(event.out_pkgs.keys()),
                        ins=', '.join(event.in_pkgs.keys()),
                        action=event.action.name,
                        to_or_by='by' if event.action == 'Replaced' else 'to',
                        is_or_are='is' if len(event.in_pkgs.keys()) == 1 else 'are'
                    )
                )

    api.current_logger().debug(message)


def filter_out_transaction_conf_pkgs(tasks, transaction_configuration):
    """
    Filter out those PES events conflicting with the higher priority transaction configuration files.

    :param tasks: A dict with three dicts holding pkgs to keep, to install and to remove
    :param transaction_configuration: RpmTransactionTasks model instance with pkgs to install, keep and REMOVE based
                                      on the user configuration files
    """
    do_not_keep = [p for p in tasks[Task.KEEP] if p in transaction_configuration.to_remove]
    do_not_install = [p for p in tasks[Task.INSTALL] if p in transaction_configuration.to_remove]
    do_not_remove = [p for p in tasks[Task.REMOVE] if p in transaction_configuration.to_install
                     or p in transaction_configuration.to_keep]

    for pkg in do_not_keep:
        # Removing a package from the to_keep dict may cause that some repositories won't get enabled
        tasks[Task.KEEP].pop(pkg)

    if do_not_install:
        for pkg in do_not_install:
            tasks[Task.INSTALL].pop(pkg)
        api.current_logger().debug('The following packages will not be installed because of the'
                                   ' /etc/leapp/transaction/to_remove transaction configuration file:'
                                   '\n- ' + '\n- '.join(sorted(do_not_install)))
    if do_not_remove:
        for pkg in do_not_remove:
            tasks[Task.REMOVE].pop(pkg)
        api.current_logger().debug('The following packages will not be removed because of the to_keep and to_install'
                                   ' transaction configuration files in /etc/leapp/transaction/:'
                                   '\n- ' + '\n- '.join(sorted(do_not_remove)))


def produce_messages(tasks):
    # Type casting to list to be Py2&Py3 compatible as on Py3 keys() returns dict_keys(), not a list
    to_install_pkgs = sorted(tasks[Task.INSTALL].keys())
    to_remove_pkgs = sorted(tasks[Task.REMOVE].keys())
    to_enable_repos = sorted(set(tasks[Task.INSTALL].values() + tasks[Task.KEEP].values()))

    if to_install_pkgs or to_remove_pkgs:
        api.produce(PESRpmTransactionTasks(to_install=to_install_pkgs,
                                           to_remove=to_remove_pkgs))

    if to_enable_repos:
        api.produce(RepositoriesSetupTasks(to_enable=to_enable_repos))



import os

from leapp.libraries.stdlib import run, api, CalledProcessError
from leapp.exceptions import StopActorExecution
from leapp.models import GrubDevice


def has_grub(blk_dev):
    """
    Check whether GRUB is present on block device
    """
    try:
        blk = os.open(blk_dev, os.O_RDONLY)
        mbr = os.read(blk, 512)
    except OSError:
        api.current_logger().warning(
            'Could not read first sector of {} in order to identify the bootloader'.format(blk_dev)
        )
        raise StopActorExecution()
    os.close(blk)
    return 'GRUB' in mbr


def blk_dev_from_partition(partition):
    """
    Find parent device of /boot partition
    """
    try:
        result = run(['lsblk', '-spnlo', 'name', partition])
    except CalledProcessError:
        api.current_logger().warning(
            'Could not get parent device of {} partition'.format(partition)
        )
        raise StopActorExecution()
    # lsblk "-s" option prints dependencies in inverse order, so the parent device will always
    # be the last or the only device.
    # Command result example:
    # 'result', {'signal': 0, 'pid': 3872, 'exit_code': 0, 'stderr': u'', 'stdout': u'/dev/vda1\n/dev/vda\n'}
    return result['stdout'].strip().split()[-1]


def get_boot_partition():
    """
    Get /boot partition
    """
    try:
        # call grub2-probe to identify /boot partition
        result = run(['grub2-probe', '--target=device', '/boot'])
    except CalledProcessError:
        api.current_logger().warning(
            'Could not get name of underlying /boot partition'
        )
        raise StopActorExecution()
    return result['stdout'].strip()


def get_grub_device():
    """
    Get block device where GRUB is located. We assume GRUB is on the same device
    as /boot partition is.

    """
    grub_dev = os.getenv('LEAPP_GRUB_DEVICE', None)
    if grub_dev:
        api.produce(GrubDevice(grub_device=grub_dev))
        return
    boot_partition = get_boot_partition()
    grub_dev = blk_dev_from_partition(boot_partition)
    if grub_dev:
        if has_grub(grub_dev):
            api.produce(GrubDevice(grub_device=grub_dev))



import os
import shutil

from leapp.actors import Actor
from leapp.models import SELinuxModules, SELinuxCustom, SELinuxRequestRPMs
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag
from leapp.libraries.stdlib import run, CalledProcessError


WORKING_DIRECTORY = '/tmp/selinux/'


class SELinuxApplyCustom(Actor):
    """
    Re-apply SELinux customizations from RHEL-7 installation

    Re-apply SELinux policy customizations (custom policy modules and changes
    introduced by semanage). Any changes (due to incompatiblity with RHEL-8
    SELinux policy) are reported to user.
    """
    name = 'selinuxapplycustom'
    consumes = (SELinuxCustom, SELinuxModules)
    produces = ()
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        # cil module files need to be extracted to disk in order to be installed

        # clear working directory
        shutil.rmtree(WORKING_DIRECTORY, ignore_errors=True)

        try:
            os.mkdir(WORKING_DIRECTORY)
        except OSError:
            self.log.warning("Failed to create working directory! Aborting.")
            return

        # import custom SElinux modules
        for semodules in self.consume(SELinuxModules):
            self.log.info("Processing custom SELinux policy modules. Count: %d.", len(semodules.modules))
            for module in semodules.modules:
                cil_filename = os.path.join(WORKING_DIRECTORY, "{}.cil".format(module.name))
                self.log.info("Installing module %s on priority %d.", module.name, module.priority)
                if module.removed:
                    self.log.warning("The following lines where removed because of incompatibility: \n%s",
                                     '\n'.join(module.removed))
                # write module content to disk
                try:
                    with open(cil_filename, 'w') as cil_file:
                        cil_file.write(module.content)
                except OSError as e:
                    self.log.warning("Error writing %s : %s", cil_filename, str(e))
                    continue

                try:
                    run(['semodule',
                         '-X', str(module.priority),
                         '-i', cil_filename
                         ]
                        )
                except CalledProcessError as e:
                    self.log.warning("Error installing module: %s", str(e.stderr))
                    # TODO - save the failed module to /etc/selinux ?
                    # currently it is still left in the old policy store
                try:
                    os.remove(cil_filename)
                except OSError as e:
                    self.log.warning("Error removing module file: %s", str(e))
        # import SELinux customizations collected by "semanage export"
        for custom in self.consume(SELinuxCustom):
            self.log.info('Importing the following SELinux customizations collected by "semanage export": \n%s',
                          '\n'.join(custom.commands))
            semanage_filename = os.path.join(WORKING_DIRECTORY, "semanage")
            # save SELinux customizations to disk
            try:
                with open(semanage_filename, 'w') as s_file:
                    s_file.write('\n'.join(custom.commands))
            except OSError as e:
                self.log.warning("Error writing SELinux customizations: %s", str(e))
            # import customizations
            try:
                run(['semanage', 'import', '-f', semanage_filename])
            except CalledProcessError as e:
                self.log.warning("Failed to import SELinux customizations: %s", str(e.stderr))
                continue
            # clean-up
            try:
                os.remove(semanage_filename)
            except OSError as e:
                self.log.warning("Failed to remove temporary file %s: %s", semanage_filename, str(e))
                continue

        # clean-up
        shutil.rmtree(WORKING_DIRECTORY, ignore_errors=True)

        # TODO - Verify that all RPM packages reqested by selinux actors are installed
        self.log.info("Verifying selinux-related RPMs requested before upgrade.")
        for rpms in self.consume(SELinuxRequestRPMs):
            self.log.info("To keep: %s \n To install: %s", ", ".join(rpms.to_keep), ", ".join(rpms.to_install))

        # TODO - summarize all changes after LEAPP team rewrites reporting
        # from leapp.reporting import Report



import os

import pytest

from leapp.snactor.fixture import current_actor_context
from leapp.models import SELinuxModule, SELinuxModules, SELinuxCustom, SELinuxFacts, SELinuxRequestRPMs
from leapp.libraries.stdlib import api, run, CalledProcessError
from leapp.reporting import Report

TEST_MODULES = [
    ["400", "mock1"],
    ["99", "mock1"],
    ["200", "mock1"],
    ["400", "mock2"],
    ["999", "mock3"],
    ["400", "permissive_abrt_t"]
]

SEMANAGE_COMMANDS = [
    ['fcontext', '-t', 'ganesha_var_run_t', "'/ganesha(/.*)?'"],
    ['fcontext', '-t', 'httpd_sys_content_t', "'/web(/.*)?'"],
    ['port', '-t', 'http_port_t', '-p', 'udp', '81']
]


def _run_cmd(cmd, logmsg="", split=True):
    try:
        return run(cmd, split=split).get("stdout", "")
    except CalledProcessError as e:
        if logmsg:
            api.current_logger().warning("%s: %s", logmsg, str(e.stderr))
    return None


def find_module_semodule(semodule_lfull, name, priority):
    return next((line for line in semodule_lfull if (name in line and priority in line)), None)


def find_semanage_rule(rules, rule):
    return next((r for r in rules if all(word in r for word in rule)), None)


@pytest.fixture(scope="function")
def destructive_selinux_teardown():
    # actor introduces changes to the system, therefore only teardown is needed
    yield

    for priority, module in TEST_MODULES:
        _run_cmd(["semodule", "-X", priority, "-r", module],
                 "Error removing module {} after testing".format(module))

    for command in SEMANAGE_COMMANDS[1:]:
        _run_cmd(["semanage", command[0], "-d"] + [x.strip('"\'') for x in command[1:]],
                 "Failed to remove SELinux customizations after testing")

    _run_cmd(["semanage", SEMANAGE_COMMANDS[0][0], "-d"] + SEMANAGE_COMMANDS[0][1:],
             "Failed to remove SELinux customizations after testing")


@pytest.mark.skipif(os.getenv("DESTRUCTIVE_TESTING", False) in [False, "0"],
                    reason='Test disabled by default because it would modify the system')
def test_SELinuxApplyCustom(current_actor_context, destructive_selinux_teardown):

    semodule_list = [SELinuxModule(name=module, priority=int(prio),
                                   content="(allow domain proc_type (file (getattr open read)))", removed=[])
                     for (prio, module) in TEST_MODULES]

    commands = [" ".join([c[0], "-a"] + c[1:]) for c in SEMANAGE_COMMANDS[1:]]
    semanage_removed = [" ".join([SEMANAGE_COMMANDS[0][0], "-a"] + SEMANAGE_COMMANDS[0][1:])]

    current_actor_context.feed(SELinuxModules(modules=semodule_list))
    current_actor_context.feed(SELinuxCustom(commands=commands, removed=semanage_removed))
    current_actor_context.run()

    semodule_lfull = _run_cmd(["semodule", "-lfull"],
                              "Error listing selinux modules")
    semanage_export = _run_cmd(["semanage", "export"],
                               "Error listing selinux customizations")

    # check that all reported modules where introduced to the system
    for priority, name in TEST_MODULES:
        if priority not in ('100', '200'):
            assert find_module_semodule(semodule_lfull, name, priority)
    # check that all valid commands where introduced to the system
    for command in SEMANAGE_COMMANDS[1:-1]:
        assert find_semanage_rule(semanage_export, command)



from leapp.actors import Actor
from leapp.libraries.actor.peseventsscanner import pes_events_scanner
from leapp.models import (
    InstalledRedHatSignedRPM,
    PESRpmTransactionTasks,
    RepositoriesMap,
    RepositoriesBlacklisted,
    RepositoriesSetupTasks,
    RpmTransactionTasks,
)
from leapp.reporting import Report
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class PesEventsScanner(Actor):
    """
    Provides data about packages events from Package Evolution Service.

    After collecting data from a provided JSON file containing Package Evolution Service events, a
    message with relevant data will be produced to help DNF Upgrade transaction calculation.
    """

    name = 'pes_events_scanner'
    consumes = (InstalledRedHatSignedRPM, RepositoriesBlacklisted, RepositoriesMap, RpmTransactionTasks)
    produces = (PESRpmTransactionTasks, RepositoriesSetupTasks, Report)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        pes_events_scanner('/etc/leapp/files', 'pes-events.json')



from leapp.actors import Actor
from leapp.libraries.actor.checkkerneldrivers import (
    check_drivers,
    get_present_drivers,
    get_removed_drivers,
)
from leapp.models import PCIDevices
from leapp.reporting import Report, create_report
from leapp import reporting
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckKernelDrivers(Actor):
    """
    Actor checks if any loaded RHEL7 kernel driver is missing in the RHEL8.
    If yes, the upgrade process will be inhibited.

    Inhibition is done because missing kernel driver on the RHEL8 system may
    mean that the hardware using such driver would not work on the RHEL8.

    Note:
     - List of kernel drivers missing on the RHEL8 system is located in the
        files/removed_drivers.txt file.
    """
    name = 'check_kernel_drivers'
    consumes = (PCIDevices,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        removed_file = 'files/removed_drivers.txt'
        conflicting = check_drivers(get_removed_drivers(removed_file), get_present_drivers())

        if conflicting:
            title = ('Detected loaded kernel drivers which have been removed '
                     'in RHEL 8. Upgrade cannot proceed.')
            URL = ('https://access.redhat.com/documentation/en-us/red_hat_enterprise_linux/8/html-single/'
                   'considerations_in_adopting_rhel_8/index#removed-device-drivers_hardware-enablement')
            summary = ('Support for the following RHEL 7 '
                       'device drivers has been removed in RHEL 8: \n     - {}'
                       '\nPlease see {} for details.'.format('\n     - '.join(conflicting), URL))
            remediation = ('Please disable detected kernel drivers in '
                           'order to proceed with the upgrade process using the rmmod or modprobe -r.')
            create_report([
                reporting.Title(title),
                reporting.Summary(summary),
                reporting.Severity(reporting.Severity.HIGH),
                reporting.Tags([reporting.Tags.KERNEL, reporting.Tags.DRIVERS]),
                reporting.Flags([reporting.Flags.INHIBITOR]),
                reporting.Remediation(hint=remediation)
            ] + [reporting.RelatedResource('kernel-driver', kd) for kd in conflicting])



from leapp.actors import Actor
from leapp.libraries.actor import checknonmountboots390
from leapp.models import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckNonMountBootS390(Actor):
    """
    Inhibits on s390 when /boot is NOT on a separate partition.

    Due to some problems, if /boot is not on a separate partition, leapp is deleting the content of /boot.
    To avoid this from happening, we are inhibiting the upgrade process until this problem has been solved.
    """

    name = 'check_non_mount_boot_s390'
    consumes = ()
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        checknonmountboots390.perform_check()



import re

from leapp.libraries.stdlib import api
from leapp.models import CPUInfo

RE_MACHINE_TYPE = re.compile(r'^processor.*\smachine\s*=\s*([0-9]+)')


def _get_cpuinfo():
    """Return lines from /proc/cpuinfo."""
    # Expecting the file exists on earch system, skipping any check
    with open('/proc/cpuinfo', 'rb') as fp:
        return fp.readlines()


def process():
    cpuinfo = CPUInfo()

    machine_types = [RE_MACHINE_TYPE.findall(line) for line in _get_cpuinfo()]
    machine_types = [i[0] for i in machine_types if i]
    if machine_types:
        # machine type should be same for all found cpus
        cpuinfo.machine_type = int(machine_types[0])
    api.produce(cpuinfo)



from leapp.actors import Actor
from leapp.libraries.actor import scankernel
from leapp.models import InstalledTargetKernelVersion, TransactionCompleted
from leapp.tags import IPUWorkflowTag, RPMUpgradePhaseTag


class ScanInstalledTargetKernelVersion(Actor):
    """
    Scan for the version of the newly installed kernel

    This actor will query rpm for all kernel packages and reports the first
    matching el8 kernel RPM. In case the RHEL Real Time has been detected on
    the original system, the kernel-rt rpm is searched. If the rpm is missing,
    fallback for standard kernel RPM.
    """

    name = 'scan_installed_target_kernel_version'
    consumes = (TransactionCompleted,)
    produces = (InstalledTargetKernelVersion,)
    tags = (RPMUpgradePhaseTag, IPUWorkflowTag)

    def process(self):
        scankernel.process()



from leapp.actors import Actor
from leapp.libraries.actor import restrictedpcisscanner
from leapp.models import RestrictedPCIDevices
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class RestrictedPCIsScanner(Actor):
    """
    Provides data about restricted (unsupported or/and unavailable) devices.

    The data sources priority set up is as following:
    1. The actor tries to get the data from /etc/leapp/files
    2. Then from the online service
    """

    name = "restricted_pcis_scanner"
    consumes = ()
    produces = (RestrictedPCIDevices,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        restrictedpcisscanner.produce_restricted_pcis()



import logging

import pytest

from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import cpu
from leapp.libraries.common import testutils
from leapp.libraries.common.config import architecture
from leapp.libraries.common.testutils import CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.models import CPUInfo


def test_non_ibmz_arch(monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(architecture.ARCH_X86_64))
    monkeypatch.setattr(reporting, "create_report", testutils.create_report_mocked())
    cpu.process()
    assert not reporting.create_report.called


def test_ibmz_arch_missing_cpuinfo(monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(architecture.ARCH_S390X))
    monkeypatch.setattr(reporting, "create_report", testutils.create_report_mocked())
    monkeypatch.setattr(api, 'consume', lambda x: iter([]))
    with pytest.raises(StopActorExecutionError):
        cpu.process()
    assert not reporting.create_report.called


def test_ibmz_cpu_supported(monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(architecture.ARCH_S390X))
    monkeypatch.setattr(reporting, "create_report", testutils.create_report_mocked())
    for sup_arch in cpu.SUPPORTED_MACHINE_TYPES:
        monkeypatch.setattr(api, 'consume', lambda x: iter([CPUInfo(machine_type=sup_arch)]))
        cpu.process()
        assert not reporting.create_report.called


def test_ibmz_cpu_unsupported(monkeypatch):
    title_msg = 'The processor is not supported by the target system.'
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(architecture.ARCH_S390X))
    monkeypatch.setattr(api, 'consume', lambda x: iter([CPUInfo(machine_type=666)]))
    monkeypatch.setattr(reporting, "create_report", testutils.create_report_mocked())
    cpu.process()
    assert reporting.create_report.called
    assert title_msg == reporting.create_report.report_fields['title']
    assert reporting.Flags.INHIBITOR in reporting.create_report.report_fields['flags']


def test_ibmz_cpu_is_empty(monkeypatch, caplog):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(architecture.ARCH_S390X))
    monkeypatch.setattr(reporting, "create_report", testutils.create_report_mocked())
    monkeypatch.setattr(api, 'consume', lambda x: iter([CPUInfo(machine_type=None)]))
    with caplog.at_level(logging.DEBUG):
        cpu.process()
    assert 'The machine (CPU) type is empty.' in caplog.text



"""
TODOs:
1. consider the idea to compare data sources timestamp in order to decide
   if the source is actual or not
"""

from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api
from leapp.models import PCIDevices, RestrictedPCIDevices
from leapp.reporting import create_report


def render_report(
    restricted_driver_names_on_host,
    restricted_pci_ids_on_host,
    restricted_devices_drivers,
    restricted_devices_pcis,
    inhibit_upgrade=False,
):
    """
    Render the report for restricted PCI devices in use.

    :param inhibit_upgrade: if True, the report will have Inhibitor flag and
        the upgrade will be inhibited.
    """
    # Prefilter needed data for the report
    unavailable_driver_names = tuple(
        driver
        for driver in restricted_driver_names_on_host
        if restricted_devices_drivers[driver].available_rhel8 == 0
    )
    unavailable_pci_ids = tuple(
        pci_id
        for pci_id in restricted_pci_ids_on_host
        if restricted_devices_pcis[pci_id].available_rhel8 == 0
    )
    unsupported_driver_names = tuple(
        driver
        for driver in restricted_driver_names_on_host
        if restricted_devices_drivers[driver].supported_rhel8 == 0
    )
    unsupported_pci_ids = tuple(
        pci_id
        for pci_id in restricted_pci_ids_on_host
        if restricted_devices_pcis[pci_id].supported_rhel8 == 0
    )

    # Render the report entities
    title = "Detected PCI drivers that are restricted on RHEL8."

    # Prepare summary
    summary_partial = "The following drivers are restricted on RHEL 8 system:"
    resources = []
    # Process unavailable devices
    if unavailable_driver_names or unavailable_pci_ids:
        summary_partial += "\n\t Unavailable in RHEL8:"
        if unavailable_driver_names:
            summary_partial += "\n\t\t- driver names:\n"
            summary_partial += "\t\t\t- " + "\n\t\t\t- ".join(
                unavailable_driver_names
            )
            resources += [reporting.RelatedResource('kernel-driver', i) for i in unavailable_driver_names]
        if unavailable_pci_ids:
            summary_partial += "\n\t\t- pci ids:\n"
            summary_partial += "\t\t\t- " + "\n\t\t\t- ".join(
                unavailable_pci_ids
            )

    # Process unsupported devices
    if unsupported_driver_names or unsupported_pci_ids:
        summary_partial += "\n\t Unsupported in RHEL8:"
        if unsupported_driver_names:
            summary_partial += "\n\t\t- driver names:\n"
            summary_partial += "\t\t\t- " + "\n\t\t\t- ".join(
                unsupported_driver_names
            )
            resources += [reporting.RelatedResource('kernel-driver', i) for i in unsupported_driver_names]
        if unsupported_pci_ids:
            summary_partial += "\n\t\t- pci ids:\n"
            summary_partial += "\t\t\t- " + "\n\t\t\t- ".join(
                unsupported_pci_ids
            )

    reports = [
        reporting.Title(title),
        reporting.Summary(summary_partial),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Tags([reporting.Tags.KERNEL, reporting.Tags.DRIVERS]),
    ]

    if resources:
        reports += resources
    if inhibit_upgrade:
        reports += [reporting.Flags([reporting.Flags.INHIBITOR])]
    return reports


def create_dict_lookup(driver_names, key):
    """
    Creates a dict lookup from the list of Models.

    :param driver_names: List[leapp.models.Model[K, V]]
    :param key: a key value, that will be used as a primary key to generate a
        result Dict
    :return: Dict[leapp.models.Model[K,V].key, leapp.models.Model[K,V]]
    :raises ValueError: if there are duplicates of leapp.models.Model[K,V].key
    :raises AttributeError: if key not in K (Model attributes)

    Example:
    >>> create_dict_lookup([Model(a=1, b=2), Model(a=2, b=3)], key="b") == {
    >>>     2: Model(a=1, b=2),
    >>>     3: Model(a=2, b=3),
    >>> }
    """
    lookup = {getattr(item, key): item for item in driver_names}
    if len(lookup) != len(driver_names):
        raise ValueError("A duplicated key(s) found")
    return lookup


def checkpcidrivers_main():
    """Main entrypoint of the CheckPCIDrivers Actor."""
    try:
        pci_devs = next(api.consume(PCIDevices))
        restricted_pci_devs = next(api.consume(RestrictedPCIDevices))
    except StopIteration:
        raise StopActorExecutionError(
            message=(
                "At least one of the needed messages is empty. "
                "Required messages are PCIDevices, RestrictedPCIDevices."
            )
        )
    else:
        # get set of drivers-names/pci-ids presented on host
        driver_names_on_host = set(dev.driver for dev in pci_devs.devices)
        pci_ids_on_host = set(dev.pci_id for dev in pci_devs.devices)

        # get restricted driver_names and pci_ids dict lookups
        try:
            restricted_devices_drivers = create_dict_lookup(
                restricted_pci_devs.driver_names, key="driver_name"
            )
            restricted_devices_pcis = create_dict_lookup(
                restricted_pci_devs.pci_ids, key="pci_id"
            )
        except (AttributeError, ValueError) as err:
            raise StopActorExecutionError(message=str(err))
        # get set of restricted driver-names/pci-ids presented on host
        restricted_driver_names_on_host = driver_names_on_host & set(
            restricted_devices_drivers
        )
        restricted_pci_ids_on_host = pci_ids_on_host & set(
            restricted_devices_pcis
        )

        # check if at least one driver on host is not available on RHEL8
        if any(
            restricted_devices_drivers[driver_name].available_rhel8 == 0
            for driver_name in restricted_driver_names_on_host
        ) or any(
            restricted_devices_pcis[pci].available_rhel8 == 0
            for pci in restricted_pci_ids_on_host
        ):
            api.current_logger().critical(
                "Some of the host drivers are unavailable on the RHEL 8 "
                "system. Inhibiting the upgrade...",
            )
            create_report(
                render_report(
                    restricted_driver_names_on_host,
                    restricted_pci_ids_on_host,
                    restricted_devices_drivers,
                    restricted_devices_pcis,
                    inhibit_upgrade=True,
                )
            )
        # check if at least one driver on host is not supported on RHEL8
        elif any(
            restricted_devices_drivers[driver_name].supported_rhel8 == 0
            for driver_name in restricted_driver_names_on_host
        ) or any(
            restricted_devices_pcis[pci].supported_rhel8 == 0
            for pci in restricted_pci_ids_on_host
        ):
            api.current_logger().warning(
                "Some of the host drivers are unsupported on the RHEL 8 "
                "system. Warning the user...",
            )
            create_report(
                render_report(
                    restricted_driver_names_on_host,
                    restricted_pci_ids_on_host,
                    restricted_devices_drivers,
                    restricted_devices_pcis,
                )
            )



import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import initrdinclude
from leapp.libraries.stdlib import api, CalledProcessError
from leapp.libraries.common.testutils import CurrentActorMocked, logger_mocked
from leapp.models import InitrdIncludes, InstalledTargetKernelVersion


INCLUDES1 = ["/file1", "/file2", "/dir/ect/ory/file3"]
INCLUDES2 = ["/file4", "/file5"]
KERNEL_VERSION = "4.18.0"


def raise_call_error(args=None):
    raise CalledProcessError(
        message='A Leapp Command Error occured.',
        command=args,
        result={'signal': None, 'exit_code': 1, 'pid': 0, 'stdout': 'fake', 'stderr': 'fake'})


class RunMocked(object):
    def __init__(self, raise_err=False):
        self.called = 0
        self.args = []
        self.raise_err = raise_err

    def __call__(self, args):
        self.called += 1
        self.args = args
        if self.raise_err:
            raise_call_error(args)


def test_no_includes(monkeypatch):
    run_mocked = RunMocked()
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=[]))
    monkeypatch.setattr(api, 'current_logger', logger_mocked())
    monkeypatch.setattr(initrdinclude, 'run', run_mocked)

    initrdinclude.process()
    assert "No additional files required to add into the initrd." in api.current_logger.dbgmsg
    assert not run_mocked.called


def test_no_kernel_version(monkeypatch):
    run_mocked = RunMocked()
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(
        msgs=[InitrdIncludes(files=INCLUDES1), InitrdIncludes(files=INCLUDES2)]))
    monkeypatch.setattr(initrdinclude, 'run', run_mocked)

    with pytest.raises(StopActorExecutionError) as e:
        initrdinclude.process()
    assert 'Cannot get version of the installed RHEL-8 kernel' in str(e)
    assert not run_mocked.called


def test_dracut_fail(monkeypatch):
    run_mocked = RunMocked(raise_err=True)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(
        msgs=[InitrdIncludes(files=INCLUDES1), InitrdIncludes(files=INCLUDES2),
              InstalledTargetKernelVersion(version=KERNEL_VERSION)]))
    monkeypatch.setattr(initrdinclude, 'run', run_mocked)

    with pytest.raises(StopActorExecutionError) as e:
        initrdinclude.process()
    assert 'Cannot regenerate dracut image' in str(e)
    assert run_mocked.called


def test_flawless(monkeypatch):
    run_mocked = RunMocked()
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(
        msgs=[InitrdIncludes(files=INCLUDES1), InitrdIncludes(files=INCLUDES2),
              InstalledTargetKernelVersion(version=KERNEL_VERSION)]))
    monkeypatch.setattr(initrdinclude, 'run', run_mocked)

    initrdinclude.process()
    assert run_mocked.called
    for f in INCLUDES1 + INCLUDES2:
        assert (f in arg for arg in run_mocked.args)



from leapp.actors import Actor
from leapp.models import ActiveKernelModulesFacts
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag
from leapp import reporting
from leapp.reporting import Report, create_report


class CheckBtrfs(Actor):
    """
    Check if Btrfs filesystem is in use. If yes, inhibit the upgrade process.

    Btrfs filesystem was introduced as Technology Preview with initial releases of RHEL 6 and 7. It
    was deprecated on versions 6.6 and 7.4 and will not be present in next major version.
    """

    name = 'check_btrfs'
    consumes = (ActiveKernelModulesFacts,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        for fact in self.consume(ActiveKernelModulesFacts):
            for active_module in fact.kernel_modules:
                if active_module.filename == 'btrfs':
                    create_report([
                        reporting.Title('Btrfs has been removed from RHEL8'),
                        reporting.Summary(
                            'The Btrfs file system was introduced as Technology Preview with the '
                            'initial release of Red Hat Enterprise Linux 6 and Red Hat Enterprise Linux 7. As of '
                            'versions 6.6 and 7.4 this technology has been deprecated and removed in RHEL8.'
                        ),
                        reporting.ExternalLink(
                            title='Considerations in adopting RHEL 8 - btrfs has been removed.',
                            url='https://access.redhat.com/documentation/en-us/red_hat_enterprise_linux/8/html/considerations_in_adopting_rhel_8/file-systems-and-storage_considerations-in-adopting-rhel-8#btrfs-has-been-removed_file-systems-and-storage'  # noqa: E501; pylint: disable=line-too-long
                        ),
                        reporting.Severity(reporting.Severity.HIGH),
                        reporting.Flags([reporting.Flags.INHIBITOR]),
                        reporting.Tags([reporting.Tags.FILESYSTEM]),
                        reporting.RelatedResource('kernel-driver', 'btrfs')
                    ])
                    break
